# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


class Boundary:
    """A Boundary as a collection of prescribed degrees of freedom (numbered
    coordinate components of a field at points of a mesh).

    Parameters
    ----------
    field : felupe.Field
        Field on wich the boundary is created.
    name : str, optional (default is "default")
        Name of the boundary.
    fx : float or callable, optional
        Mask-function for x-component of mesh-points which returns `True` at points
        on which the boundary will be applied (default is ``np.isnan``). If a float is
        passed, this is transformed to ``lambda x: np.isclose(x, fx)``.
    fy : float or callable, optional
        Mask-function for y-component of mesh-points which returns `True` at points
        on which the boundary will be applied (default is ``np.isnan``). If a float is
        passed, this is transformed to ``lambda y: np.isclose(y, fy)``.
    fz : float or callable, optional
        Mask-function for z-component of mesh-points which returns `True` at points
        on which the boundary will be applied (default is ``np.isnan``). If a float is
        passed, this is transformed to ``lambda z: np.isclose(z, fz)``.
    value : ndarray or float, optional
        Value(s) of the selected (prescribed) degrees of freedom (default is 0.0).
    skip : None or tuple of bool or int, optional
        A tuple to define which axes of the selected points should be skipped, i.e.
        not prescribed (default is ``None`` and will be set to ``(False, False, False)``
        if ``mask=None``).
    mask : ndarray
        Boolean mask for the prescribed degrees of freedom. If a mask is passed, ``fx``,
        ``fy`` and ``fz`` are ignored. However, ``skip`` is still applied on the mask.
    mode : string, optional
        A string which defines the logical operation for the selected points per
        axis (default is `or`).

    Attributes
    ----------
    mask : ndarray
        1d- or 2d-boolean mask array for the prescribed degrees of freedom.
    dof : ndarray
        1d-array of ints which contains the prescribed degrees of freedom.
    points : ndarray
        1d-array of ints which contains the point ids on which one or more degrees of
        freedom are prescribed.
    value : ndarray or float
        Value of the selected (prescribed) degrees of freedom.

    Examples
    --------
    A boundary condition prescribes values for chosen degrees of freedom of a given
    field (**not** a field container). This is demonstrated for a plane-strain
    vector field on a quad-mesh of a circle.

    >>> import felupe as fem

    >>> mesh = fem.Circle(radius=1, n=6)
    >>> x, y = mesh.points.T
    >>> region = fem.RegionQuad(mesh)
    >>> displacement = fem.FieldPlaneStrain(region, dim=2)
    >>> field = fem.FieldContainer([displacement])

    A boundary on the displacement field which prescribes all components of the field
    on the outermost left point of the circle is created. The easiest way is to pass the
    desired value to ``fx``. The same result is obtained if a callable function is
    passed to ``fx``.

    >>> left = fem.Boundary(displacement, fx=x.min())
    >>> left = fem.Boundary(displacement, fx=lambda x: np.isclose(x, x.min()))

    >>> plotter = mesh.plot(off_screen=True)
    >>> plotter.add_points(
    >>>     np.pad(mesh.points[left.points], ((0, 0), (0, 1))),
    >>>     point_size=20,
    >>>     color="red",
    >>> )
    >>> img = plotter.screenshot("boundary_left.png", transparent_background=True)

    ..  image:: images/boundary_left.png

    If ``fx`` and ``fy`` are given, the masks are combined by *logical-or*.

    >>> axes = fem.Boundary(displacement, fx=0, fy=0, mode="or")

    ..  image:: images/boundary_axes.png

    This may be changed to *logical-and* if desired.

    >>> center = fem.Boundary(displacement, fx=0, fy=0, mode="and")

    ..  image:: images/boundary_center.png

    For the most-general case, a user-defined boolean mask for the selection of the
    mesh-points is provided. While the two upper methods are useful to select
    points separated per point-coordinates, providing a mask is more flexible as
    it may involve all three coordinates (or any other quantities of interest).

    >>> mask = np.logical_and(np.isclose(x**2 + y**2, 1), x <= 0)
    >>> surface = fem.Boundary(displacement, mask=mask)

    ..  image:: images/boundary_surface.png

    The application of a new mask allows to change the selected points of an existing
    boundary condition.

    >>> new_mask = np.logical_and(mask, y >= 0)
    >>> surface.apply_mask(new_mask)

    ..  image:: images/boundary_new_surface.png

    A boundary condition may be skipped on given axes, i.e. if only the x-components
    of a field should be prescribed on the selected points, then the y-axis must
    be skipped.

    >>> axes_x = fem.Boundary(displacement, fx=0, fy=0, skip=(False, True))

    Values for the prescribed degress of freedom are either applied during creation
    or by the update-method.

    >>> left = fem.Boundary(displacement, fx=x.min(), value=-0.2)
    >>> left.update(-0.3)

    Sometimes it is useful to create a boundary with all axes skipped. This
    boundary has no prescribed degrees of freedom and hence, is without effect.
    However, it may still be used in a characteristic job for the boundary to be
    tracked.

    See Also
    --------
    felupe.CharacteristicCurve : A job with a boundary to be tracked.
    felupe.dof.partition : Partition degrees of freedom into prescribed and active dof.
    felupe.dof.apply : Apply prescribed values for a list of boundaries.

    """

    def __init__(
        self,
        field,
        name="default",
        fx=np.isnan,
        fy=np.isnan,
        fz=np.isnan,
        value=0.0,
        skip=None,
        mask=None,
        mode="or",
    ):
        mesh = field.region.mesh

        self.field = field
        self.dim = field.dim
        self.name = name
        self.value = value
        self.skip = skip
        self.mode = mode

        # check if callable
        _fx = fx if callable(fx) else lambda x: np.isclose(x, fx)
        _fy = fy if callable(fy) else lambda y: np.isclose(y, fy)
        _fz = fz if callable(fz) else lambda z: np.isclose(z, fz)

        self.fun = [_fx, _fy, _fz][: mesh.dim]

        if self.skip is None:
            self.skip = (False, False, False)

        if mask is None:
            self.skip = np.array(self.skip).astype(int)[: self.dim]

            # apply functions on the points per coordinate
            # fx(x), fy(y), fz(z) and create a mask for each coordinate
            masks = [f(x) for f, x in zip(self.fun, mesh.points.T)]

            # select the logical combination function "or" or "and"
            combine = {"or": np.logical_or, "and": np.logical_and}[self.mode]
            mask = combine.reduce(masks)

        self.apply_mask(mask)

    def apply_mask(self, mask):
        "Apply a boolean mask to the boundary."

        # reshape the mask
        npoints = self.field.region.mesh.npoints
        self.mask = mask.reshape(npoints, -1)

        # expand point-based to dof-based mask
        if self.mask.shape[1] == 1:
            self.mask = np.tile(self.mask, self.dim)

            # check if some axes should be skipped
            if True in self.skip:
                # exclude mask from axes which should be skipped
                self.mask[:, np.where(self.skip)[0]] = False

        elif self.mask.shape[1] == self.dim:
            self.skip = None

        else:
            raise ValueError(
                " ".join(
                    [
                        "The given mask is of wrong shape:",
                        "for point-based masks it must be",
                        "(field.region.mesh.npoints, 1)",
                        "and for dof-based masks this must be",
                        "(field.region.mesh.npoints, field.dim).",
                    ]
                )
            )

        self.dof = self.field.indices.dof[self.mask]
        self.points = np.arange(self.field.region.mesh.npoints)[self.mask.any(axis=1)]

    def update(self, value):
        "Update the value of the boundary in-place."

        self.value = value



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..field._axi import FieldAxisymmetric
from ..field._base import Field
from ._cartesian import IntegralFormCartesian


class IntegralFormAxisymmetric(IntegralFormCartesian):
    r"""An Integral Form for axisymmetric fields.

    Notes
    -----
    Axisymmetric scenarios are modeled with a 2D-mesh and consequently, a 2D element
    formulation. The rotation axis is chosen along the global X-axis
    :math:`(X,Y,Z) \widehat{=} (Z,R,\varphi)`. The 3x3 deformation gradient consists of
    an in-plane 2x2 sub-matrix and one additional entry for the out-of-plane stretch
    which is equal to the ratio of deformed and undeformed radius.

    ..  math::

        \boldsymbol{F} = \begin{bmatrix}
                \boldsymbol{F}_{(2D)} & \boldsymbol{0} \\
                \boldsymbol{0}^T & \frac{r}{R}
            \end{bmatrix}

    The variation of the deformation gradient consists of both in- and out-of-plane
    contributions.

    ..  math::

        \delta \boldsymbol{F}_{(2D)} = \delta \frac{
                \partial \boldsymbol{u}}{\partial \boldsymbol{X}
            } \qquad \text{and} \qquad \delta \left(\frac{r}{R}\right)
            = \frac{\delta u_r}{R}

    Again, the internal virtual work leads to two seperate terms.

    ..  math::

        -\delta W_{int} = \int_V \boldsymbol{P} : \delta \boldsymbol{F} \ dV
            = \int_V \boldsymbol{P}_{(2D)} : \delta \boldsymbol{F}_{(2D)} \ dV
            + \int_V \frac{P_{33}}{R} : \delta u_r \ dV

    The differential volume is further expressed as a product of the differential
    in-plane area and the differential arc length. The arc length integral is finally
    pre-evaluated.

    ..  math::

        \int_V dV = \int_{\varphi=0}^{2\pi} \int_A R\ dA\ d\varphi = 2\pi \int_A R\ dA

    Inserting the differential volume integral into the expression of internal virtual
    work, this leads to:

    ..  math::

        -\delta W_{int} = 2\pi \int_A \boldsymbol{P}_{(2D)}
            : \delta \boldsymbol{F}_{(2D)} \ R \ dA
            + 2\pi \int_A P_{33} : \delta u_r \ dA

    A Linearization of the internal virtual work expression gives four terms.

    ..  math::

        -\Delta \delta W_{int} &= \Delta_{(2D)} \delta_{(2D)} W_{int}
            + \Delta_{33} \delta_{(2D)} W_{int} + \Delta_{(2D)} \delta_{33} W_{int}
            + \Delta_{33} \delta_{33} W_{int}

        -\Delta_{(2D)} \delta_{(2D)} W_{int} &= 2\pi \int_A
            \delta \boldsymbol{F}_{(2D)} : \mathbb{A}_{(2D),(2D)} :
                \Delta \boldsymbol{F}_{(2D)} \ R \ dA

        -\Delta_{33} \delta_{(2D)} W_{int} &= 2\pi \int_A
            \delta \boldsymbol{F}_{(2D)} : \mathbb{A}_{(2D),33} : \Delta u_r \ dA

        -\Delta_{(2D)} \delta_{33} W_{int} &= 2\pi \int_A
            \delta u_r : \mathbb{A}_{33,(2D)} : \Delta \boldsymbol{F}_{(2D)} \ dA

        -\Delta_{33} \delta_{33} W_{int} &= 2\pi \int_A
            \delta u_r : \frac{\mathbb{A}_{33,33}}{R} : \Delta u_r \ dA

    with

    ..  math::

        \mathbb{A}_{(2D),(2D)} &= \frac{
            \partial \psi}{\partial \boldsymbol{F}_{(2D)} \partial \boldsymbol{F}_{(2D)}
        }

        \mathbb{A}_{(2D),33} &= \frac{
            \partial \psi}{\partial \boldsymbol{F}_{(2D)} \partial F^3_{\hphantom{3}3}}
            \left ( = \mathbb{A}_{33,(2D)} \right )

        \mathbb{A}_{33,33} &= \frac{\partial \psi}{F^3_{\hphantom{3}3}
            \partial F^3_{\hphantom{3}3}}
    """

    def __init__(self, fun, v, dV, u=None, grad_v=True, grad_u=True):
        R = v.radius
        self.dV = 2 * np.pi * R * dV

        if u is None:
            if isinstance(v, FieldAxisymmetric):
                self.mode = 1

                if grad_v:
                    fun_2d = fun[:-1, :-1]
                    fun_zz = fun[(-1,), (-1,)] / R
                else:
                    fun_2d = fun[:-1]
                    fun_zz = fun[-1].reshape(1, *fun[-1].shape) / R

                form_a = IntegralFormCartesian(fun_2d, v, self.dV, grad_v=grad_v)
                form_b = IntegralFormCartesian(fun_zz, v.scalar, self.dV)

                self.forms = [form_a, form_b]

            else:
                self.mode = 10

                form_a = IntegralFormCartesian(fun, v, self.dV, grad_v=False)
                self.forms = [
                    form_a,
                ]

        else:
            if isinstance(v, FieldAxisymmetric) and isinstance(u, FieldAxisymmetric):
                self.mode = 2

                if grad_v and grad_u:
                    form_aa = IntegralFormCartesian(
                        fun[:-1, :-1, :-1, :-1], v, self.dV, u, True, True
                    )
                    form_bb = IntegralFormCartesian(
                        fun[-1, -1, -1, -1] / R**2,
                        v.scalar,
                        self.dV,
                        u.scalar,
                        False,
                        False,
                    )
                    form_ba = IntegralFormCartesian(
                        fun[-1, -1, :-1, :-1] / R, v.scalar, self.dV, u, False, True
                    )
                    form_ab = IntegralFormCartesian(
                        fun[:-1, :-1, -1, -1] / R, v, self.dV, u.scalar, True, False
                    )

                if not grad_v and grad_u:
                    form_aa = IntegralFormCartesian(
                        fun[:-1, :-1, :-1], v, self.dV, u, False, True
                    )
                    form_bb = IntegralFormCartesian(
                        fun[-1, -1, -1] / R**2,
                        v.scalar,
                        self.dV,
                        u.scalar,
                        False,
                        False,
                    )
                    form_ba = IntegralFormCartesian(
                        fun[-1, :-1, :-1] / R, v.scalar, self.dV, u, False, True
                    )
                    form_ab = IntegralFormCartesian(
                        fun[:-1, -1, -1] / R, v, self.dV, u.scalar, False, False
                    )

                self.forms = [form_aa, form_bb, form_ba, form_ab]

            elif isinstance(v, FieldAxisymmetric) and isinstance(u, Field):
                self.mode = 30

                form_a = IntegralFormCartesian(
                    fun[:-1, :-1], v, self.dV, u, True, False
                )
                form_b = IntegralFormCartesian(
                    fun[-1, -1] / R, v.scalar, self.dV, u, False, False
                )

                self.forms = [form_a, form_b]

            elif isinstance(v, Field) and isinstance(u, Field):
                self.mode = 40

                form_a = IntegralFormCartesian(fun, v, self.dV, u, False, False)

                self.forms = [
                    form_a,
                ]

    def integrate(self, parallel=False):
        values = [form.integrate(parallel=parallel) for form in self.forms]

        if self.mode == 1:
            values[0] += np.pad(values[1], ((0, 0), (1, 0), (0, 0)))
            val = values[0]

        if self.mode == 30:
            if len(values[0].shape) > 4:
                values[0] = values[0][:, :, 0, 0]
            if len(values[1].shape) > 4:
                values[1] = values[1][:, :, 0, 0]
            a, b, e = values[1].shape
            values[1] = values[1].reshape(a, 1, b, e)

            values[0] += np.pad(values[1], ((0, 0), (1, 0), (0, 0), (0, 0)))
            val = values[0]

        elif self.mode == 2:
            a, b, e = values[1].shape
            values[1] = values[1].reshape(a, 1, b, 1, e)
            values[1] = np.pad(values[1], ((0, 0), (1, 0), (0, 0), (1, 0), (0, 0)))

            a, b, i, e = values[2].shape
            values[2] = values[2].reshape(a, 1, b, i, e)
            values[2] = np.pad(values[2], ((0, 0), (1, 0), (0, 0), (0, 0), (0, 0)))

            a, i, b, e = values[3].shape
            values[3] = values[3].reshape(a, i, b, 1, e)
            values[3] = np.pad(values[3], ((0, 0), (0, 0), (0, 0), (1, 0), (0, 0)))

            for i in range(1, len(values)):
                values[0] += values[i]

            val = values[0]

        elif self.mode == 10 or self.mode == 40:
            val = values[0]

        return val

    def assemble(self, values=None, parallel=False):
        if values is None:
            values = self.integrate(parallel=parallel)
        return self.forms[0].assemble(values)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..math import cross
from ..mesh import Mesh
from ._region import Region


def boundary_cells_quad(mesh):
    "Convert the cells array of a quad mesh into a boundary cells array."

    # edges (boundary) of a quad
    i = [3, 1, 0, 2]
    j = [0, 2, 1, 3]

    cells_faces = np.dstack(
        (
            mesh.cells[:, i],
            mesh.cells[:, j],
        )
    )

    # complementary edges for the creation of "boundary" quads
    # (rotated quads with 1st edge as n-th edge of one original quad)
    t = [1, 0, 3, 2]
    k = np.array(i)[t]
    m = np.array(j)[t]

    cells = np.dstack(
        (
            cells_faces,
            mesh.cells[:, k],
            mesh.cells[:, m],
        )
    )

    return cells, cells_faces


def boundary_cells_quad8(mesh):
    "Convert the cells array of a quadratic quad mesh into a boundary cells array."

    cells_quad, cells_faces_quad = boundary_cells_quad(mesh)

    # midpoints on edges (boundary) of a quadratic quad
    m = [7, 5, 4, 6]

    cells_faces = np.dstack(
        (
            cells_faces_quad,
            mesh.cells[:, m],
        )
    )

    # complementary midpoints of edges for the creation of "boundary" quadratic quads
    # (rotated quads with 1st edge as n-th edge of one original quad)
    n = [5, 6, 7, 4]
    p = [4, 5, 6, 7]
    q = [6, 7, 4, 5]

    cells = np.dstack(
        (
            cells_quad,
            mesh.cells[:, m],
            mesh.cells[:, n],
            mesh.cells[:, p],
            mesh.cells[:, q],
        )
    )

    return cells, cells_faces


def boundary_cells_quad9(mesh):
    "Convert the cells array of a bi-quadratic quad mesh into a boundary cells array."

    cells_quad8, cells_faces_quad8 = boundary_cells_quad8(mesh)

    # midpoints on edges (boundary) of a quadratic quad
    r = [8, 8, 8, 8]

    cells = np.dstack(
        (
            cells_quad8,
            mesh.cells[:, r],
        )
    )

    return cells, cells_faces_quad8


def boundary_cells_hexahedron(mesh):
    "Convert the cells array of a hex mesh into a boundary cells array."

    # faces (boundary) of a hexahedron
    i = [0, 2, 1, 3, 0, 5]
    j = [3, 1, 0, 2, 1, 4]
    k = [7, 5, 4, 6, 2, 7]
    r = [4, 6, 5, 7, 3, 6]

    cells_faces = np.dstack(
        (
            mesh.cells[:, i],
            mesh.cells[:, j],
            mesh.cells[:, k],
            mesh.cells[:, r],
        )
    )

    # complementary faces for the creation of "boundary" hexahedrons
    # (6 rotated hexahedrons with 1st face as n-th face of
    #  one original hexahedron)
    m = [1, 3, 2, 0, 4, 1]
    n = [2, 0, 3, 1, 5, 0]
    p = [6, 4, 7, 5, 6, 3]
    q = [5, 7, 6, 4, 7, 2]

    cells = np.dstack(
        (
            cells_faces,
            mesh.cells[:, m],
            mesh.cells[:, n],
            mesh.cells[:, p],
            mesh.cells[:, q],
        )
    )

    return cells, cells_faces


def boundary_cells_hexahedron20(mesh):
    "Convert the cells array of a quadratic hex mesh into a boundary cells array."

    cells_hexahedron, cells_faces_hexahedron = boundary_cells_hexahedron(mesh)

    # midpoints on edges of faces (boundary) of a hexahedron
    i = [11, 9, 8, 10, 8, 12]
    j = [19, 17, 16, 18, 9, 15]
    k = [15, 13, 12, 14, 10, 14]
    v = [16, 18, 17, 19, 11, 13]

    cells_faces = np.dstack(
        (
            cells_faces_hexahedron,
            mesh.cells[:, i],
            mesh.cells[:, j],
            mesh.cells[:, k],
            mesh.cells[:, v],
        )
    )

    # complementary faces for the creation of "boundary" hexahedrons
    # (6 rotated hexahedrons with 1st face as n-th face of
    #  one original hexahedron)
    m = [9, 11, 10, 8, 12, 8]
    n = [18, 16, 19, 17, 13, 11]
    p = [13, 15, 14, 12, 14, 10]
    q = [17, 19, 18, 16, 15, 9]

    r = [8, 10, 9, 11, 16, 17]
    s = [10, 8, 11, 9, 17, 16]
    t = [14, 12, 15, 13, 18, 19]
    u = [12, 14, 13, 15, 19, 18]

    cells = np.dstack(
        (
            cells_hexahedron,
            mesh.cells[:, i],
            mesh.cells[:, j],
            mesh.cells[:, k],
            mesh.cells[:, v],
            mesh.cells[:, m],
            mesh.cells[:, n],
            mesh.cells[:, p],
            mesh.cells[:, q],
            mesh.cells[:, r],
            mesh.cells[:, s],
            mesh.cells[:, t],
            mesh.cells[:, u],
        )
    )

    return cells, cells_faces


def boundary_cells_hexahedron27(mesh):
    "Convert the cells array of a bi-quadratic hex mesh into a boundary cells array."

    cells_hexahedron20, cells_faces_hexahedron20 = boundary_cells_hexahedron20(mesh)

    # midpoints of faces (boundary) of a hexahedron
    i = [20, 21, 22, 23, 24, 25]

    cells_faces = np.dstack(
        (
            cells_faces_hexahedron20,
            mesh.cells[:, i],
        )
    )

    # complementary faces for the creation of "boundary" hexahedrons
    # (6 rotated hexahedrons with 1st face as n-th face of
    #  one original hexahedron)
    j = [22, 23, 21, 20, 20, 21]
    k = [23, 22, 20, 21, 21, 20]
    r = [24, 25, 24, 25, 22, 23]
    m = [25, 24, 25, 24, 23, 22]
    n = [20, 21, 22, 23, 24, 25]
    p = [21, 20, 23, 22, 25, 24]
    q = [26, 26, 26, 26, 26, 26]

    cells = np.dstack(
        (
            cells_hexahedron20,
            mesh.cells[:, j],
            mesh.cells[:, k],
            mesh.cells[:, r],
            mesh.cells[:, m],
            mesh.cells[:, n],
            mesh.cells[:, p],
            mesh.cells[:, q],
        )
    )

    return cells, cells_faces


class RegionBoundary(Region):
    r"""
    A numeric boundary-region as a combination of a mesh, an element and a
    numeric integration scheme (quadrature rule).

    Parameters
    ----------
    mesh : Mesh
        A mesh with points and cells.
    element : Element
        The finite element formulation to be applied on the cells.
    quadrature: Quadrature
        An element-compatible numeric integration scheme with points and weights.
    grad : bool, optional
        A flag to invoke gradient evaluation (default is True).
    only_surface: bool, optional
        A flag to use only the enclosing outline of the region (default is True).
    mask: ndarray or None, optional
        A boolean array to select a specific set of points (default is None).
    ensure_3d : bool, optional
        A flag to enforce 3d area normal vectors.

    Attributes
    ----------
    mesh : Mesh
        A mesh with points and cells.
    element : Finite element
        The finite element formulation to be applied on the cells.
    quadrature: Quadrature scheme
        An element-compatible numeric integration scheme with points and weights.
    h : ndarray
        Element shape function array ``h_aq`` of shape function ``a`` evaluated at
        quadrature point ``q``.
    dhdr : ndarray
        Partial derivative of element shape function array ``dhdr_aJq`` with shape
        function ``a`` w.r.t. natural element coordinate ``J`` evaluated at quadrature
        point ``q`` for every cell ``c`` (geometric gradient or **Jacobian**
        transformation between ``X`` and ``r``).
    dXdr : ndarray
        Geometric gradient ``dXdr_IJqc`` as partial derivative of undeformed coordinate
        ``I`` w.r.t. natural element coordinate ``J`` evaluated at quadrature point
        ``q`` for every cell ``c`` (geometric gradient or **Jacobian** transformation
        between ``X`` and ``r``).
    drdX : ndarray
        Inverse of dXdr.
    dA : ndarray
        Numeric *Differential area vectors*.
    normals : ndarray
        Area unit normal vectors.
    dV : ndarray
        Numeric *Differential volume element* as norm of *Differential area vectors*.
    dhdX : ndarray
        Partial derivative of element shape functions ``dhdX_aJqc`` of shape function
        ``a`` w.r.t. undeformed coordinate ``J`` evaluated at quadrature point ``q`` for
        every cell ``c``.

    Notes
    -----
    The gradients of the element shape functions w.r.t the undeformed coordinates are
    evaluated at all integration points of each cell in the region if the optional
    gradient argument is ``True``.

    .. math::

       \frac{\partial X_I}{\partial r_J} &= \hat{X}_{aI} \frac{
               \partial h_a}{\partial r_J
           }

       \frac{\partial h_a}{\partial X_J} &= \frac{\partial h_a}{\partial r_I}
       \frac{\partial r_I}{\partial X_J}

       dV &= \det\left(\frac{\partial X_I}{\partial r_J}\right) w

    Examples
    --------
    >>> import felupe as fem

    >>> mesh = fem.Rectangle(n=(3, 2))
    >>> element = fem.Quad()
    >>> quadrature = fem.GaussLegendreBoundary(order=1, dim=2)

    >>> region = fem.RegionBoundary(mesh, element, quadrature)
    >>> region
    <felupe Region object>
      Element formulation: Quad
      Quadrature rule: GaussLegendreBoundary
      Gradient evaluated: True

    The numeric differential area vectors are the products of the cofactors of the
    geometric gradient :math:`\partial X_I / \partial r_J` and the weights `w` of the
    quadrature points. The differential area vectors array is of shape
    ``(nquadraturepoints, ndim, nboundarycells)``.

    >>> region.dA
    array([[[ 0.  , -0.5 ,  0.  ,  0.5 ,  0.  ,  0.  ],
            [ 0.  , -0.5 ,  0.  ,  0.5 ,  0.  ,  0.  ]],

           [[-0.25, -0.  , -0.25, -0.  ,  0.25,  0.25],
            [-0.25, -0.  , -0.25, -0.  ,  0.25,  0.25]]])

    In a boundary region, the numeric differential volumes are the magnitudes of the
    differential area vectors. For a quad mesh, the boundary cell volumes are the edge
    lengths.

    >>> region.dV.sum(axis=0)
    array([0.5, 1. , 0.5, 1. , 0.5, 0.5])

    Unit normal vectors are obtained by the ratio of the differential area vectors and
    the differential volumes.

    >>> region.dA / region.dV  ## this is equal to ``region.normals``
    array([[[ 0., -1.,  0.,  1.,  0.,  0.],
            [ 0., -1.,  0.,  1.,  0.,  0.]],

           [[-1., -0., -1., -0.,  1.,  1.],
            [-1., -0., -1., -0.,  1.,  1.]]])

    The partial derivative of the first element shape function w.r.t. the undeformed
    coordinates evaluated at the second integration point of the last element of the
    region:

    >>> region.dhdX[0, :, 1, -1]
    array([2.        , 0.21132487])
    """

    def __init__(
        self,
        mesh,
        element,
        quadrature,
        grad=True,
        only_surface=True,
        mask=None,
        ensure_3d=False,
    ):
        self.only_surface = only_surface
        self.mask = mask
        self.ensure_3d = ensure_3d

        if mesh.cell_type == "quad":
            cells, cells_faces = boundary_cells_quad(mesh)
        elif mesh.cell_type == "quad8":
            cells, cells_faces = boundary_cells_quad8(mesh)
        elif mesh.cell_type == "quad9":
            cells, cells_faces = boundary_cells_quad9(mesh)
        elif mesh.cell_type == "hexahedron":
            cells, cells_faces = boundary_cells_hexahedron(mesh)
        elif mesh.cell_type == "hexahedron20":
            cells, cells_faces = boundary_cells_hexahedron20(mesh)
        elif mesh.cell_type == "hexahedron27":
            cells, cells_faces = boundary_cells_hexahedron27(mesh)
        else:
            raise NotImplementedError("Cell type not supported.")

        cells_faces = cells_faces.reshape(-1, cells_faces.shape[-1])
        cells = cells.reshape(-1, cells.shape[-1])

        if self.only_surface:
            # sort faces, get indices of unique faces and counts
            cells_faces_sorted = np.sort(cells_faces, axis=1)
            cells_faces_unique, index, counts = np.unique(
                cells_faces_sorted, True, False, True, 0
            )

            self._index = index
            self._mask = counts == 1
        else:
            self._index = np.arange(len(cells_faces))
            self._mask = np.ones_like(self._index, dtype=bool)

        self._selection = self._index[self._mask]

        # merge with point-mask
        if mask is not None:
            point_selection = np.arange(len(mesh.points))[mask]
            self._selection = self._selection[
                np.all(np.isin(cells_faces[self._selection], point_selection), axis=1)
            ]

        # get cell-faces and cells on boundary (unique cell-faces with one count)
        cells_on_boundary = cells[self._selection]

        # create mesh on boundary
        mesh_boundary_cells = mesh.copy()
        mesh_boundary_cells.update(cells=cells_on_boundary)
        self.mesh = mesh_boundary_cells
        self.mesh.cells_faces = cells_faces[self._selection]

        # init region and faces
        super().__init__(mesh_boundary_cells, element, quadrature, grad=grad)

        if grad:
            self.dA, self.dV, self.normals = self._init_faces()

    def _init_faces(self):
        "Initialize (norm of) face normals of cells."

        if (
            self.mesh.cell_type == "quad"
            or self.mesh.cell_type == "quad8"
            or self.mesh.cell_type == "quad9"
        ):
            dA_1 = self.dXdr[:, 0][::-1]
            dA_1[0] = -dA_1[0]

        elif (
            self.mesh.cell_type == "hexahedron"
            or self.mesh.cell_type == "hexahedron20"
            or self.mesh.cell_type == "hexahedron27"
        ):
            dA_1 = cross(self.dXdr[:, 0], self.dXdr[:, 1])

        dA = -dA_1 * self.quadrature.weights.reshape(-1, 1)

        # norm and unit normal vector
        dV = np.linalg.norm(dA, axis=0)
        normals = dA / dV

        if self.ensure_3d:
            if dA.shape[0] == 2:
                dA = np.pad(dA, ((0, 1), (0, 0), (0, 0)))
                normals = np.pad(normals, ((0, 1), (0, 0), (0, 0)))

        return dA, dV, normals

    def mesh_faces(self):
        "Return a Mesh with face-cells on the selected boundary."

        face_type = {
            "quad": "line",
            "hexahedron": "quad",
            "quad8": "line3",
            "quad9": "line3",
        }[self.mesh.cell_type]
        return Mesh(self.mesh.points, self.mesh.cells_faces, face_type)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._line_rectangle_cube import cube_hexa, line_line, rectangle_quad
from ._mesh import Mesh
from ._tools import concatenate


class Line(Mesh):
    "A line shaped 1d-mesh with lines between ``a`` and ``b`` with ``n`` points."

    def __init__(self, a=0, b=1, n=2):
        self.a = a
        self.b = b
        self.n = n

        points, cells, cell_type = line_line(a, b, n)

        super().__init__(points, cells, cell_type)


class Rectangle(Mesh):
    """A rectangular 2d-mesh with quads between ``a`` and ``b`` with ``n``
    points per axis."""

    def __init__(self, a=(0, 0), b=(1, 1), n=(2, 2)):
        self.a = a
        self.b = b
        self.n = n

        points, cells, cell_type = rectangle_quad(a, b, n)

        super().__init__(points, cells, cell_type)


class Cube(Mesh):
    """A cube shaped 3d-mesh with hexahedrons between ``a`` and ``b`` with ``n``
    points per axis."""

    def __init__(self, a=(0, 0, 0), b=(1, 1, 1), n=(2, 2, 2)):
        self.a = a
        self.b = b
        self.n = n

        points, cells, cell_type = cube_hexa(a, b, n)

        super().__init__(points, cells, cell_type)


class Grid(Mesh):
    """A grid shaped 3d-mesh with hexahedrons. Basically a wrapper for
    :func:`numpy.meshgrid` with  default ``indexing="ij"``."""

    def __init__(self, *xi, indexing="ij", **kwargs):
        shape = np.array([len(x) for x in xi])
        n = shape if len(shape) > 1 else shape[0]

        M = [None, line_line, rectangle_quad, cube_hexa][len(xi)]
        points, cells, cell_type = M(n=n)

        super().__init__(points, cells, cell_type)

        self.points = (
            np.vstack(
                [g.T.ravel() for g in np.meshgrid(*xi, indexing=indexing, **kwargs)]
            )
            .astype(float)
            .T
        )


class RectangleArbitraryOrderQuad(Mesh):
    """A rectangular 2d-mesh with arbitrary order quads between ``a`` and ``b`` with
    ``n`` points per axis."""

    def __init__(self, a=(0, 0), b=(1, 1), order=2):
        yv, xv = np.meshgrid(
            np.linspace(a[1], b[1], order + 1),
            np.linspace(a[0], b[0], order + 1),
            indexing="ij",
        )

        points = np.vstack((xv.flatten(), yv.flatten())).T

        # search vertices
        xmin = min(points[:, 0])
        ymin = min(points[:, 1])

        xmax = max(points[:, 0])
        ymax = max(points[:, 1])

        def search_vertice(p, x, y):
            return np.where(np.logical_and(p[:, 0] == x, p[:, 1] == y))[0][0]

        def search_edge(p, a, x):
            return np.where(p[:, a] == x)[0][1:-1]

        v1 = search_vertice(points, xmin, ymin)
        v2 = search_vertice(points, xmax, ymin)
        v3 = search_vertice(points, xmax, ymax)
        v4 = search_vertice(points, xmin, ymax)

        vertices = [v1, v2, v3, v4]

        mask1 = np.ones_like(points[:, 0], dtype=bool)
        mask1[vertices] = 0
        # points_no_verts = points[mask1]

        e1 = search_edge(points, 1, ymin)
        e2 = search_edge(points, 0, xmax)
        e3 = search_edge(points, 1, ymax)
        e4 = search_edge(points, 0, xmin)

        edges = np.hstack((e1, e2, e3, e4))

        mask2 = np.ones_like(points[:, 0], dtype=bool)
        mask2[np.hstack((vertices, edges))] = 0
        # points_no_verts_edges = points[mask2]

        face = np.arange(len(points))[mask2]

        cells = np.hstack((vertices, edges, face)).reshape(1, -1)

        super().__init__(points, cells, cell_type="VTK_LAGRANGE_QUADRILATERAL")


class CubeArbitraryOrderHexahedron(Mesh):
    """A cube shaped 3d-mesh with arbitrary order hexahedrons between ``a`` and ``b``
    with ``n`` points per axis."""

    def __init__(self, a=(0, 0, 0), b=(1, 1, 1), order=2):
        zv, yv, xv = np.meshgrid(
            np.linspace(a[2], b[2], order + 1),
            np.linspace(a[1], b[1], order + 1),
            np.linspace(a[0], b[0], order + 1),
            indexing="ij",
        )

        points = np.vstack((xv.flatten(), yv.flatten(), zv.flatten())).T

        # search vertices
        xmin = min(points[:, 0])
        ymin = min(points[:, 1])
        zmin = min(points[:, 2])

        xmax = max(points[:, 0])
        ymax = max(points[:, 1])
        zmax = max(points[:, 2])

        def search_vertice(p, x, y, z):
            return np.where(
                np.logical_and(np.logical_and(p[:, 0] == x, p[:, 1] == y), p[:, 2] == z)
            )[0][0]

        def search_edge(p, a, b, x, y):
            return np.where(np.logical_and(p[:, a] == x, p[:, b] == y))[0][1:-1]

        def search_face(p, a, x, vertices, edges):
            face = np.where(points[:, a] == x)[0]
            mask = np.zeros_like(p[:, 0], dtype=bool)
            mask[face] = 1
            mask[np.hstack((vertices, edges))] = 0
            return np.arange(len(p[:, 0]))[mask]

        v1 = search_vertice(points, xmin, ymin, zmin)
        v2 = search_vertice(points, xmax, ymin, zmin)
        v3 = search_vertice(points, xmax, ymax, zmin)
        v4 = search_vertice(points, xmin, ymax, zmin)

        v5 = search_vertice(points, xmin, ymin, zmax)
        v6 = search_vertice(points, xmax, ymin, zmax)
        v7 = search_vertice(points, xmax, ymax, zmax)
        v8 = search_vertice(points, xmin, ymax, zmax)

        vertices = [v1, v2, v3, v4, v5, v6, v7, v8]

        mask1 = np.ones_like(points[:, 0], dtype=bool)
        mask1[vertices] = 0
        # points_no_verts = points[mask1]

        e1 = search_edge(points, 1, 2, ymin, zmin)
        e2 = search_edge(points, 0, 2, xmax, zmin)
        e3 = search_edge(points, 1, 2, ymax, zmin)
        e4 = search_edge(points, 0, 2, xmin, zmin)

        e5 = search_edge(points, 1, 2, ymin, zmax)
        e6 = search_edge(points, 0, 2, xmax, zmax)
        e7 = search_edge(points, 1, 2, ymax, zmax)
        e8 = search_edge(points, 0, 2, xmin, zmax)

        e9 = search_edge(points, 0, 1, xmin, ymin)
        e10 = search_edge(points, 0, 1, xmax, ymin)
        e11 = search_edge(points, 0, 1, xmin, ymax)
        e12 = search_edge(points, 0, 1, xmax, ymax)

        edges = np.hstack((e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e12))

        mask2 = np.ones_like(points[:, 0], dtype=bool)
        mask2[np.hstack((vertices, edges))] = 0
        # points_no_verts_edges = points[mask2]

        f1 = search_face(points, 0, xmin, vertices, edges)
        f2 = search_face(points, 0, xmax, vertices, edges)
        f3 = search_face(points, 1, ymin, vertices, edges)
        f4 = search_face(points, 1, ymax, vertices, edges)
        f5 = search_face(points, 2, zmin, vertices, edges)
        f6 = search_face(points, 2, zmax, vertices, edges)

        faces = np.hstack((f1, f2, f3, f4, f5, f6))

        mask3 = np.ones_like(points[:, 0], dtype=bool)
        mask3[np.hstack((vertices, edges, faces))] = 0
        volume = np.arange(len(points))[mask3]

        cells = np.hstack((vertices, edges, faces, volume)).reshape(1, -1)

        super().__init__(points, cells, cell_type="VTK_LAGRANGE_HEXAHEDRON")


class Circle(Mesh):
    """A circular shaped 2d-mesh with quads and ``n`` points on the circumferential
    edge of a 45-degree section. 90-degree ``sections`` are placed at given angles in
    degree."""

    def __init__(
        self,
        radius=1,
        centerpoint=[0, 0],
        n=2,
        sections=[0, 90, 180, 270],
        value=0.15,
        exponent=2,
        decimals=10,
    ):
        rect = Rectangle(b=(0.5, 0.5), n=(n, n))
        right = rect.points[:, 0] == 0.5

        rect = rect.add_runouts(values=[-value], axis=0, exponent=exponent)
        rect.points[:, 0] = rect.points[:, 1].reshape(n, n, order="F").ravel()

        line = Line(n=n)
        phi = np.linspace(np.pi / 4, 0, n)

        bottom = line.copy(points=rect.points[right][::-1])
        top = line.copy(points=np.vstack([np.cos(phi), np.sin(phi)]).T)
        face = bottom.fill_between(top, n=n)

        quarter = concatenate([face, face.mirror(normal=[-1, 1]), rect])
        circle = concatenate([quarter.rotate(alpha, 2) for alpha in sections]).sweep(
            decimals=decimals
        )

        circle.points *= radius
        circle.points += np.array(centerpoint)

        super().__init__(
            points=circle.points, cells=circle.cells, cell_type=circle.cell_type
        )


class Triangle(Mesh):
    """A triangular shaped 2d-mesh with quads and ``n`` points at the edges of the three
    sub-quadrilaterals."""

    def __init__(
        self,
        a=(0, 0),
        b=(1, 0),
        c=(0, 1),
        n=2,
        decimals=10,
    ):
        a = np.asarray(a)
        b = np.asarray(b)
        c = np.asarray(c)

        sections = []

        centerpoint = (a + b + c) / 3
        centerpoints = {"ab": (a + b) / 2, "bc": (b + c) / 2, "ac": (a + c) / 2}

        line = Line(n=n)

        # section (connected to point) a
        x1 = np.linspace(a[0], centerpoints["ac"][0], n)
        y1 = np.linspace(a[1], centerpoints["ac"][1], n)

        left = line.copy(points=np.vstack([x1, y1]).T)

        x2 = np.linspace(centerpoints["ab"][0], centerpoint[0], n)
        y2 = np.linspace(centerpoints["ab"][1], centerpoint[1], n)

        middle = line.copy(points=np.vstack([x2, y2]).T)

        sections.append(middle.fill_between(left, n=n))

        # section (connected to point) b
        x3 = np.linspace(b[0], centerpoints["bc"][0], n)
        y3 = np.linspace(b[1], centerpoints["bc"][1], n)

        right = line.copy(points=np.vstack([x3, y3]).T)

        sections.append(right.fill_between(middle, n=n))

        # section (connected to point) c
        x4 = np.linspace(centerpoints["ac"][0], c[0], n)
        y4 = np.linspace(centerpoints["ac"][1], c[1], n)

        top = line.copy(points=np.vstack([x4, y4]).T)

        x5 = np.linspace(centerpoint[0], centerpoints["bc"][0], n)
        y5 = np.linspace(centerpoint[1], centerpoints["bc"][1], n)

        bottom = line.copy(points=np.vstack([x5, y5]).T)

        sections.append(bottom.fill_between(top, n=n))

        # combine sections
        triangle = concatenate(sections).sweep(decimals=decimals)

        super().__init__(
            points=triangle.points, cells=triangle.cells, cell_type=triangle.cell_type
        )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import warnings

import numpy as np

from ..math import det, inv


class Region:
    r"""
    A numeric region as a combination of a mesh, an element and a numeric integration
    scheme (quadrature rule).

    Parameters
    ----------
    mesh : Mesh
        A mesh with points and cells.
    element : Element
        The finite element formulation to be applied on the cells.
    quadrature: Quadrature
        An element-compatible numeric integration scheme with points and weights.
    grad : bool, optional
        A flag to invoke gradient evaluation (default is True).

    Attributes
    ----------
    mesh : Mesh
        A mesh with points and cells.
    element : Finite element
        The finite element formulation to be applied on the cells.
    quadrature: Quadrature scheme
        An element-compatible numeric integration scheme with points and weights.
    h : ndarray
        Element shape function array ``h_aq`` of shape function ``a`` evaluated at
        quadrature point ``q``.
    dhdr : ndarray
        Partial derivative of element shape function array ``dhdr_aJq`` with shape
        function ``a`` w.r.t. natural element coordinate ``J`` evaluated at quadrature
        point ``q`` for every cell ``c`` (geometric gradient or **Jacobian**
        transformation between ``X`` and ``r``).
    dXdr : ndarray
        Geometric gradient ``dXdr_IJqc`` as partial derivative of undeformed coordinate
        ``I`` w.r.t. natural element coordinate ``J`` evaluated at quadrature point
        ``q`` for every cell ``c`` (geometric gradient or **Jacobian** transformation
        between ``X`` and ``r``).
    drdX : ndarray
        Inverse of dXdr.
    dV : ndarray
        Numeric *differential volume element* as product of determinant of geometric
        gradient  ``dV_qc = det(dXdr)_qc w_q`` and quadrature weight ``w_q``, evaluated
        at quadrature point ``q`` for every cell ``c``.
    dhdX : ndarray
        Partial derivative of element shape functions ``dhdX_aJqc`` of shape function
        ``a`` w.r.t. undeformed coordinate ``J`` evaluated at quadrature point ``q`` for
        every cell ``c``.

    Notes
    -----
    The gradients of the element shape functions w.r.t the undeformed coordinates are
    evaluated at all integration points of each cell in the region if the optional
    gradient argument is ``True``.

    .. math::

       \frac{\partial X_I}{\partial r_J} &= \hat{X}_{aI} \frac{
               \partial h_a}{\partial r_J
           }

       \frac{\partial h_a}{\partial X_J} &= \frac{\partial h_a}{\partial r_I}
       \frac{\partial r_I}{\partial X_J}

       dV &= \det\left(\frac{\partial X_I}{\partial r_J}\right) w

    Examples
    --------
    >>> import felupe as fem

    >>> mesh = fem.Rectangle(n=(3, 2))
    >>> element = fem.Quad()
    >>> quadrature = fem.GaussLegendre(order=1, dim=2)

    >>> region = fem.Region(mesh, element, quadrature)
    >>> region
    <felupe Region object>
      Element formulation: Quad
      Quadrature rule: GaussLegendre
      Gradient evaluated: True

    The numeric differential volumes are the products of the determinant of the
    geometric gradient :math:`\partial X_I / \partial r_J` and the weights `w` of the
    quadrature points. The differential volume array is of shape
    ``(nquadraturepoints, ncells)``.

    >>> region.dV
    array([[0.125, 0.125],
           [0.125, 0.125],
           [0.125, 0.125],
           [0.125, 0.125]])

    Cell-volumes may be obtained by cell-wise sums of the differential volumes located
    at the quadrature points.

    >>> region.dV.sum(axis=0)
    array([0.5, 0.5])

    The partial derivative of the first element shape function w.r.t. the undeformed
    coordinates evaluated at the second integration point of the last element of the
    region:

    >>> region.dhdX[0, :, 1, -1]
    array([-1.57735027, -0.21132487])

    """

    def __init__(self, mesh, element, quadrature, grad=True):
        self.evaluate_gradient = grad
        self.reload(mesh=mesh, element=element, quadrature=quadrature)

    def reload(self, mesh=None, element=None, quadrature=None):
        """Reload the numeric region.

        Parameters
        ----------
        mesh : Mesh or None, optional
            A mesh with points and cells (default is None).
        element : Element or None, optional
            The finite element formulation to be applied on the cells (default is None).
        quadrature: Quadrature or None, optional
            An element-compatible numeric integration scheme with points and weights
            (default is None).
        """

        if mesh is not None:
            self.mesh = mesh

        if element is not None:
            self.element = element

        if quadrature is not None:
            self.quadrature = quadrature

        # element shape function
        self.element.h = np.array(
            [self.element.function(q) for q in self.quadrature.points]
        ).T
        self.h = np.tile(np.expand_dims(self.element.h, -1), self.mesh.ncells)

        # partial derivative of element shape function
        self.element.dhdr = np.array(
            [self.element.gradient(q) for q in self.quadrature.points]
        ).transpose(1, 2, 0)
        self.dhdr = np.tile(np.expand_dims(self.element.dhdr, -1), self.mesh.ncells)

        if self.evaluate_gradient:
            # geometric gradient
            self.dXdr = np.einsum(
                "caI,aJqc->IJqc", self.mesh.points[self.mesh.cells], self.dhdr
            )

            # inverse of dXdr
            self.drdX = inv(self.dXdr)

            # numeric **differential volume element**
            self.dV = det(self.dXdr) * self.quadrature.weights.reshape(-1, 1)

            # check for negative **differential volume elements**
            if np.any(self.dV < 0):
                cells_negative_volume = np.where(np.any(self.dV < 0, axis=0))[0]
                message_negative_volumes = "".join(
                    [
                        f"Negative volumes for cells \n {cells_negative_volume}\n",
                        "Try ``mesh.flip(np.any(region.dV < 0, axis=0))`` ",
                        "and re-create the region.",
                    ]
                )
                warnings.warn(message_negative_volumes)

            # Partial derivative of element shape function
            # w.r.t. undeformed coordinates
            self.dhdX = np.einsum("aIqc,IJqc->aJqc", self.dhdr, self.drdX)

    def __repr__(self):
        header = "<felupe Region object>"
        element = f"  Element formulation: {type(self.element).__name__}"
        quadrature = f"  Quadrature rule: {type(self.quadrature).__name__}"
        grad = f"  Gradient evaluated: {hasattr(self, 'dV')}"

        return "\n".join([header, element, quadrature, grad])

    def plot(self, **kwargs):
        """Plot the element with point-ids and the quadrature points,
        scaled by their weights."""

        return self.quadrature.plot(plotter=self.element.plot(**kwargs), add_axes=False)

    def screenshot(
        self,
        filename=None,
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the element with the applied quadrature.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        if filename is None:
            filename = f"region-{self.element.cell_type}.png"

        return self.plot(off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._base import Element


class Tetra(Element):
    r"""A 3D tetrahedron element formulation with linear shape functions.

    Notes
    -----
    The tetrahedron element is defined by four points (0-3). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s,t)`.

    .. math::

       \boldsymbol{h}(r,s,t) = \begin{bmatrix}
               1-r-s-t \\
               r \\
               s \\
               t
           \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.Tetra().screenshot()

    ..  image:: images/tetra.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self):
        super().__init__(shape=(4, 3))
        self.points = np.array(
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float
        )
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "tetra"

    def function(self, rst):
        "Return the shape functions at given coordinates (r, s, t)."
        r, s, t = rst
        return np.array([1 - r - s - t, r, s, t])

    def gradient(self, rst):
        "Return the gradient of shape functions at given coordinates (r, s, t)."
        r, s, t = rst
        return np.array([[-1, -1, -1], [1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float)


class TetraMINI(Element):
    r"""A 3D tetrahedron element formulation with bubble-enriched linear shape
    functions.

    Notes
    -----
    The MINI tetrahedron element is defined by five points (0-4). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s,t)`.

    .. math::

       \boldsymbol{h}(r,s,t) = \begin{bmatrix}
               1-r-s-t \\
               r \\
               s \\
               t \\
               r s t (1-r-s-t)
           \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.TetraMINI().screenshot(filename="tetra-mini.png")

    ..  image:: images/tetra-mini.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self, bubble_multiplier=1.0):
        super().__init__(shape=(5, 3))
        self.points = np.array(
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1 / 3, 1 / 3, 1 / 3]],
            dtype=float,
        )
        self.cells = np.arange(len(self.points) - 1).reshape(1, -1)
        self.cell_type = "tetra"
        self.bubble_multiplier = bubble_multiplier

    def function(self, rst):
        "Return the shape functions at given coordinates (r, s, t)."
        r, s, t = rst
        a = self.bubble_multiplier
        return np.array([1 - r - s - t, r, s, t, a * r * s * t * (1 - r - s - t)])

    def gradient(self, rst):
        "Return the gradient of shape functions at given coordinates (r, s, t)."
        r, s, t = rst
        a = self.bubble_multiplier
        return np.array(
            [
                [-1, -1, -1],
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [
                    a * (s * t * (1 - r - s - t) - r * s * t),
                    a * (r * t * (1 - r - s - t) - r * s * t),
                    a * (r * s * (1 - r - s - t) - r * s * t),
                ],
            ],
            dtype=float,
        )


class QuadraticTetra(Element):
    r"""A 3D tetrahedron element formulation with quadratic shape functions.

    Notes
    -----
    The quadratic tetrahedron element is defined by ten points (0-9). The element
    includes a mid-edge point on each of the edges of the tetrahedron. The ordering of
    the ten points defining the cell is point ids (0-3,4-9) where ids 0-3 are the four
    tetra vertices; and point ids 4-9 are the mid-edge points between (0,1), (1,2),
    (2,0), (0,3), (1,3), and (2,3). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s,t)`.

    .. math::

       \boldsymbol{h}(r,s,t) = \begin{bmatrix}
               t_1 (2 t_1 - 1) \\
               t_2 (2 t_2 - 1) \\
               t_3 (2 t_3 - 1) \\
               t_4 (2 t_4 - 1) \\
               4 t_1 t_2 \\
               4 t_2 t_3 \\
               4 t_3 t_1 \\
               4 t_1 t_4 \\
               4 t_2 t_4 \\
               4 t_3 t_4
           \end{bmatrix}

    with

    .. math::

       t_1 &= 1 - r - s - t

       t_2 &= r

       t_3 &= s

       t_4 &= t

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.QuadraticTetra().screenshot()

    ..  image:: images/tetra10.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self):
        super().__init__(shape=(10, 3))
        self.points = np.zeros(self.shape)
        self.points[:4] = np.array(
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float
        )
        self.points[4] = np.mean(self.points[[0, 1]], axis=0)
        self.points[5] = np.mean(self.points[[1, 2]], axis=0)
        self.points[6] = np.mean(self.points[[2, 0]], axis=0)
        self.points[7] = np.mean(self.points[[0, 3]], axis=0)
        self.points[8] = np.mean(self.points[[1, 3]], axis=0)
        self.points[9] = np.mean(self.points[[2, 3]], axis=0)

        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "tetra10"

    def function(self, rst):
        "Return the shape functions at given coordinates (r, s, t)."
        r, s, t = rst

        t1 = 1 - r - s - t
        t2 = r
        t3 = s
        t4 = t

        h = np.array(
            [
                t1 * (2 * t1 - 1),
                t2 * (2 * t2 - 1),
                t3 * (2 * t3 - 1),
                t4 * (2 * t4 - 1),
                4 * t1 * t2,
                4 * t2 * t3,
                4 * t3 * t1,
                4 * t1 * t4,
                4 * t2 * t4,
                4 * t3 * t4,
            ]
        )

        return h

    def gradient(self, rst):
        "Return the gradient of shape functions at given coordinates (r, s, t)."
        r, s, t = rst

        t1 = 1 - r - s - t
        t2 = r
        t3 = s
        t4 = t

        dhdt = np.array(
            [
                [4 * t1 - 1, 0, 0, 0],
                [0, 4 * t2 - 1, 0, 0],
                [0, 0, 4 * t3 - 1, 0],
                [0, 0, 0, 4 * t4 - 1],
                [4 * t2, 4 * t1, 0, 0],
                [0, 4 * t3, 4 * t2, 0],
                [4 * t3, 0, 4 * t1, 0],
                [4 * t4, 0, 0, 4 * t1],
                [0, 4 * t4, 0, 4 * t2],
                [0, 0, 4 * t4, 4 * t3],
            ]
        )

        dtdr = np.array([[-1, -1, -1], [1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float)

        return np.dot(dhdt, dtdr)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


class Scheme:
    r"""A quadrature scheme with integration `points` :math:`x_q` and `weights`
    :math:`w_q`. It approximates the integral of a function over a region :math:`V` by a
    weighted sum of function values :math:`f_q = f(x_q)`, evaluated on the quadrature-
    points.

    Notes
    -----

    The approximation is given by

    ..  math::

        \int_V f(x) dV \approx \sum f(x_q) w_q

    with quadrature points :math:`x_q` and corresponding weights :math:`w_q`.
    """

    def __init__(self, points, weights):
        self.points = points
        self.weights = weights

        self.npoints, self.dim = self.points.shape

    def plot(self, plotter=None, add_axes=True, point_size=100, **kwargs):
        """Plot the quadrature points, scaled by their weights, into a (optionally
        provided) PyVista plotter.

        See Also
        --------
        felupe.Scene.plot: Plot method of a scene.
        """

        if plotter is None:
            from ..mesh import Mesh

            mesh = Mesh(self.points, np.zeros((0, 2), dtype=int), "line")
            plotter = mesh.plot(**kwargs)

        for weight, point in zip(self.weights, self.points):
            # plotter requires 3d-point coordinates
            points = np.pad([point], ((0, 0), (0, 3 - self.dim)))

            plotter.add_points(
                points=points,
                point_size=point_size * weight / self.weights.max(),
                render_points_as_spheres=True,
                opacity=0.8,
                color="grey",
            )

        if add_axes:
            plotter.add_axes(xlabel="r", ylabel="s", zlabel="t")

        return plotter

    def screenshot(
        self,
        *args,
        filename=None,
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the quadrature.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        if filename is None:
            filename = "quadrature.png"

        return self.plot(*args, off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._boundary import Boundary
from ._tools import apply, partition


def _get_first_field(field):
    "Get first field of FieldContainer."

    return field.fields[0]


def symmetry(field, axes=(True, True, True), x=0.0, y=0.0, z=0.0, bounds=None):
    """Return a dict of boundaries for the symmetry axes on the x-, y- and
    z-coordinates.

    Parameters
    ----------
    field : felupe.Field
        Field on wich the symmetry boundaries are created.
    axes : tuple of bool or int
        Flags to invoke symmetries on the x-, y- and z-axis.
    x : float, optional
        Center of the x-symmetry (default is 0.0).
    y : float, optional
        Center of the y-symmetry (default is 0.0).
    z : float, optional
        Center of the z-symmetry (default is 0.0).
    bounds : dict of felupe.Boundary, optional
        Extend a given dict of boundaries by the symmetry boundaries (default is None).

    Returns
    -------
    dict of felupe.Boundary
        New or extended dict of boundaries including symmetry boundaries.

    Notes
    -----
    The symmetry boundaries are labeled as ``"symx"``, ``"symy"`` and ``"symz"``.

    +---------------+-------------------------+--------------------------+
    | Symmetry Axis | Prescribed (Fixed) Axes |      Skip-Argument       |
    +===============+=========================+==========================+
    |       x       |           y, z          | ``(True, False, False)`` |
    +---------------+-------------------------+--------------------------+
    |       y       |           x, z          | ``(False, True, False)`` |
    +---------------+-------------------------+--------------------------+
    |       z       |           x, y          | ``(False, False, True)`` |
    +---------------+-------------------------+--------------------------+

    Examples
    --------
    The x-symmetry boundary for a symmetry on the x-axis contains all points at the
    given x-coordinate. The degrees of freedom are prescribed except for the symmetry
    x-axis.

    >>> import numpy as np
    >>> import felupe as fem

    >>> mesh = fem.Circle(radius=1, n=6, sections=[0, 270])
    >>> x, y = mesh.points.T
    >>> region = fem.RegionQuad(mesh)
    >>> displacement = fem.FieldPlaneStrain(region, dim=2)

    >>> boundaries = fem.dof.symmetry(displacement, axes=(True, False), x=0.0)

    >>> plotter = mesh.plot(off_screen=True)
    >>> plotter.add_points(
    >>>     np.pad(mesh.points[boundaries["symx"].points], ((0, 0), (0, 1))),
    >>>     point_size=20,
    >>>     color="red",
    >>> )
    >>> img = plotter.screenshot("boundary_symx.png", transparent_background=True)

    ..  image:: images/boundary_symx.png

    See Also
    --------
    felupe.Boundary : A collection of prescribed degrees of freedom.
    """

    # convert axes to array and slice by mesh dimension
    enforce = np.array(axes).astype(bool)[: field.dim]

    # invert boolean identity matrix and use its rows
    # for the skip argument (a symmetry condition on
    # axis "z" fixes all displacements u_z=0 but keeps
    # in-plane displacements active)
    skipax = ~np.eye(3).astype(bool)
    kwarglist = [
        {"fx": x, "skip": skipax[0][: field.dim]},
        {"fy": y, "skip": skipax[1][: field.dim]},
        {"fz": z, "skip": skipax[2][: field.dim]},
    ]

    if bounds is None:
        bounds = {}
    labels = ["symx", "symy", "symz"]

    # loop over symmetry conditions and add them to a new dict
    for a, (symaxis, kwargs) in enumerate(zip(enforce, kwarglist[: field.dim])):
        if symaxis:
            bounds[labels[a]] = Boundary(field, **kwargs)

    return bounds


def uniaxial(field, left=None, right=None, move=0.2, axis=0, clamped=False, sym=True):
    """Return a dict of boundaries for uniaxial loading between a left (fixed or
    symmetry face) and a right (applied) end face along a given axis with optional
    selective symmetries at the origin. Optionally, the right end face is assumed to be
    rigid (clamped) in the transversal directions perpendicular to the longitudinal
    loading direction.

    Parameters
    ----------
    field : felupe.FieldContainer
        FieldContainer on wich the symmetry boundaries are created.
    left : float or None, optional
        The position of the left end face along the given axis (default is None). If
        None, the outermost left position of the mesh-points is taken, i.e.
        ``left=field.region.mesh.points[:, axis].min()``.
    right : float or None, optional
        The position of the right end face where the longitudinal movement is applied
        along the given axis (default is None). If None, the outermost right position
        of the mesh-points is taken, i.e.
        ``right=field.region.mesh.points[:, axis].max()``.
    move : float, optional
        The value of the longitudinal displacement applied at the right end face
        (default is 0.2).
    axis : int, optional
        The longitudinal axis (default is 0).
    clamped : bool, optional
        A flag to assume the right end face to be rigid, i.e. zero
        displacements in the direction of the transversal axes are enforced (default is
        True).
    sym : bool or tuple of bool, optional
        A flag to invoke all (bool) or individual (tuple) symmetry boundaries at the
        left end face in the direction of the longitudinal axis as well as in the
        directions of the transversal axes.

    Returns
    -------
    dict of felupe.Boundary
        Dict of boundaries for a uniaxial loadcase.
    dict of ndarray
        Loadcase-related partitioned prescribed ``dof0`` and active ``dof1`` degrees of
        freedom as well as the external displacement values ``ext0`` for the prescribed
        degrees of freedom.

    Examples
    --------
    A quarter of a solid hyperelastic cube is subjected to uniaxial displacement-
    controlled compression on a rigid end face.

    >>> import felupe as fem

    >>> region = fem.RegionHexahedron(fem.Cube(a=(0, 0, 0), b=(2, 3, 1), n=(11, 16, 6)))
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])

    >>> boundaries = fem.dof.uniaxial(field, axis=2, clamped=True)[0]

    The longitudinal displacement is applied incrementally.

    >>> solid = fem.SolidBodyNearlyIncompressible(fem.NeoHooke(mu=1), field, bulk=5000)
    >>> step = fem.Step(
    >>>     items=[solid],
    >>>     ramp={boundaries["move"]: fem.math.linsteps([0, -0.3], num=5)},
    >>>     boundaries=boundaries
    >>> )

    >>> fem.Job(steps=[step]).evaluate()
    >>> img = field.screenshot("Principal Values of Logarithmic Strain")

    ..  image:: images/loadcase_ux.png

    See Also
    --------
    felupe.Boundary : A collection of prescribed degrees of freedom.
    felupe.dof.partition : Partition degrees of freedom into prescribed and active dof.
    felupe.dof.apply : Apply prescribed values for a list of boundaries.
    felupe.dof.symmetry : Return a dict of boundaries for the symmetry axes.

    """

    f = _get_first_field(field)

    fx = ["fx", "fy", "fz"][axis]

    mask = np.ones(3, dtype=bool)
    mask[axis] = False

    active = tuple(mask.astype(int))
    inactive = tuple((~mask).astype(int))

    if not hasattr(sym, "__len__"):
        sym = (sym, sym, sym)

    if right is None:
        right = f.region.mesh.points[:, axis].max()

    bounds = symmetry(f, axes=sym)

    if not sym[axis]:
        if left is None:
            left = f.region.mesh.points[:, axis].min()

        bounds["left-x"] = Boundary(f, skip=active, **{fx: left})

    if clamped:
        bounds["right"] = Boundary(f, skip=inactive, **{fx: right})

        if not sym[axis]:
            bounds["left-yz"] = Boundary(f, skip=inactive, **{fx: left})

    bounds["move"] = Boundary(f, skip=active, value=move, **{fx: right})

    dof0, dof1 = partition(field, bounds)
    ext0 = apply(field, bounds, dof0)

    return bounds, dict(dof0=dof0, dof1=dof1, ext0=ext0)


def biaxial(
    field,
    lefts=(None, None),
    rights=(None, None),
    moves=(0.2, 0.2),
    axes=(0, 1),
    clampes=(False, False),
    sym=True,
    move=None,
):
    """Return a dict of boundaries for biaxial loading between a left (applied or
    symmetry face) and a right (applied) end face along a given pair of axes with
    optional selective symmetries at the origin. Optionally, the applied end faces are
    assumed to be rigid (clamped) in the transversal directions perpendicular to the
    longitudinal loading direction.

    Parameters
    ----------
    field : felupe.FieldContainer
        FieldContainer on wich the symmetry boundaries are created.
    lefts : tuple of float or None, optional
        The position of the left end faces where the longitudinal movement is applied
        along the given axes (default is (None, None)). If an item of the tuple is None,
        the outermost left position of the mesh-points is taken, i.e.
        ``lefts=[field.region.mesh.points[:, axis].min() for axis in axes]``.
    rights : tuple of float or None, optional
        The position of the right end faces where the longitudinal movement is applied
        along the given axes (default is (None, None)). If an item of the tuple is None,
        the outermost right position of the mesh-points is taken, i.e.
        ``rights=[field.region.mesh.points[:, axis].max() for axis in axes]``.
    moves : tuple of float, optional
        The values of the longitudinal displacements applied each one half of the value
        at the left and right end faces (default is (0.2, 0.2)).
    axes : tuple of int, optional
        The pair of longitudinal axes (default is (0, 1)).
    clampes : tuple of bool, optional
        Flags to assume the applied end faces to be rigid, i.e. zero
        displacements in the direction of the transversal axes are enforced (default is
        True).
    sym : bool or tuple of bool, optional
        A flag to invoke all (bool) or individual (tuple) symmetry boundaries at the
        left end face in the directions of the longitudinal axes as well as in the
        direction of the transversal axis.

    Returns
    -------
    dict of felupe.Boundary
        Dict of boundaries for a biaxial loadcase.
    dict of ndarray
        Loadcase-related partitioned prescribed ``dof0`` and active ``dof1`` degrees of
        freedom as well as the external displacement values ``ext0`` for the prescribed
        degrees of freedom.

    Notes
    -----
    ..  warning:: Note that `clampes=(True, True)` is not a valid loadcase for a cube.
        Instead, use a shape where the clamped end faces do not share mesh-points.

    Examples
    --------
    A cross-like planar specimen of a hyperelastic solid is subjected to biaxial
    displacement-controlled tension on rigid end faces.

    >>> import numpy as np
    >>> import felupe as fem

    >>> mesh = fem.Rectangle(a=(0, 0), b=(1, 1), n=(21, 21))
    >>> x, y = mesh.points.T
    >>> points = np.arange(mesh.npoints)[np.logical_or.reduce([x <= 0.6, y <= 0.6])]
    >>> mesh.update(cells=mesh.cells[np.all(np.isin(mesh.cells, points), axis=1)])

    >>> region = fem.RegionQuad(mesh)
    >>> field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])

    >>> boundaries = fem.dof.biaxial(field, clampes=(True, True))[0]

    The longitudinal displacements are applied incrementally.

    >>> solid = fem.SolidBodyNearlyIncompressible(fem.NeoHooke(mu=1), field, bulk=5000)
    >>> step = fem.Step(
    >>>     items=[solid],
    >>>     ramp={
    >>>         boundaries["move-right-0"]: fem.math.linsteps([0, 0.1], num=5),
    >>>         boundaries["move-right-1"]: fem.math.linsteps([0, 0.1], num=5),
    >>>     },
    >>>     boundaries=boundaries
    >>> )

    >>> fem.Job(steps=[step]).evaluate()
    >>> img = field.screenshot("Principal Values of Logarithmic Strain")

    ..  image:: images/loadcase_bx.png

    Repeating the above example with ``fem.dof.biaxial(field, clampes=(False, False)``
    results in a different deformation at the end faces.

    ..  image:: images/loadcase_bx_free.png

    The biaxial load case may also invoke a planar loading, where one of the
    longitudinal axes is fixed with no displacements at the end plates. The clampling
    must at least be deactivated on the fixed longitudinal axis.

    >>> mesh = fem.Cube(n=2)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>> boundaries = fem.dof.biaxial(
    >>>     field, clampes=(True, False), moves=(0, 0), sym=False, axes=(0, 1)
    >>> )[0]
    >>> solid = fem.SolidBodyNearlyIncompressible(fem.NeoHooke(mu=1), field, bulk=5000)
    >>> step = fem.Step(
    >>>     items=[solid],
    >>>     ramp={boundaries["move-right-0"]: fem.math.linsteps([0, 0.3], num=5),},
    >>>     boundaries=boundaries
    >>> )
    >>> fem.Job(steps=[step]).evaluate()
    >>> img = field.screenshot("Principal Values of Logarithmic Strain")

    ..  image:: images/loadcase_ps.png

    See Also
    --------
    felupe.Boundary : A collection of prescribed degrees of freedom.
    felupe.dof.partition : Partition degrees of freedom into prescribed and active dof.
    felupe.dof.apply : Apply prescribed values for a list of boundaries.
    felupe.dof.symmetry : Return a dict of boundaries for the symmetry axes.

    """

    if move is not None:
        from warnings import warn

        warn(
            "The `move` argument is deprecated, use `moves=(move, move)` instead.",
            DeprecationWarning,
        )
        moves = (move, move)

    f = _get_first_field(field)

    fxyz = ["fx", "fy", "fz"]

    lefts = np.asarray(lefts)
    rights = np.asarray(rights)

    masks = [np.ones(3, dtype=bool), np.ones(3, dtype=bool)]
    for i, axis in enumerate(axes):
        masks[i][axis] = False

    actives = [tuple(mask.astype(int)) for mask in masks]
    inactives = [tuple((~mask).astype(int)) for mask in masks]

    if not hasattr(sym, "__len__"):
        sym = (sym, sym, sym)

    for i, (right, axis) in enumerate(zip(rights, axes)):
        if right is None:
            rights[i] = f.region.mesh.points[:, axis].max()

    bounds = symmetry(f, axes=sym)

    for i, (left, axis, active, move) in enumerate(zip(lefts, axes, actives, moves)):
        if not sym[axis]:
            if left is None:
                lefts[i] = f.region.mesh.points[:, axis].min()

            fx = fxyz[axis]
            bounds[f"move-left-{axis}"] = Boundary(
                f, skip=active, value=-move, **{fx: lefts[i]}
            )

    for i, (left, right, axis, active, inactive, move, clamped) in enumerate(
        zip(lefts, rights, axes, actives, inactives, moves, clampes)
    ):
        fx = fxyz[axis]
        if clamped:
            bounds[f"right-{axis}"] = Boundary(f, skip=inactive, **{fx: right})

            if not sym[axis]:
                bounds[f"left-{axis}"] = Boundary(f, skip=inactive, **{fx: left})

        bounds[f"move-right-{axis}"] = Boundary(
            f, skip=active, value=move, **{fx: right}
        )

    dof0, dof1 = partition(field, bounds)
    ext0 = apply(field, bounds, dof0)

    return bounds, dict(dof0=dof0, dof1=dof1, ext0=ext0)


def shear(
    field,
    bottom=None,
    top=None,
    moves=(0.2, 0.0, 0.0),
    axes=(0, 1),
    sym=True,
    move=None,
    axis_shear=None,
    axis_compression=None,
    compression=None,
):
    """Return a dict of boundaries for shear loading with optional combined compression
    between a rigid bottom and a rigid top end face along a given pair of axes. The
    first axis is the direction of shear and the second axis the direction of
    compression. The bottom face remains fixed while the shear is applied at the top
    face. Optionally, a symmetry boundary condition in the thickness direction at
    the origin may be added.

    Parameters
    ----------
    field : felupe.FieldContainer
        FieldContainer on wich the symmetry boundaries are created.
    bottom : float or None, optional
        The position of the bottom end face (default is None). If None, the outermost
        bottom position of the mesh-points is taken, i.e.
        ``bottom=[field.region.mesh.points[:, axis].min() for axis in axes]``.
    top : float or None, optional
        The position of the top end face (default is None). If None, the outermost
        top position of the mesh-points is taken, i.e.
        ``top=[field.region.mesh.points[:, axis].min() for axis in axes]``.
    moves : tuple of float, optional
        The values of the displacements applied on the end faces (default is
        (0.2, 0.0, 0.0)). The first item is the shear displacement applied on the top
        end face. The second and third items refer to the tension/compression
        displacements. The second item is applied on the bottom and the third item on
        the top end face.
    axes : tuple of int, optional
        The pair of axes: the first item is the axis of shear and the second item is the
        axis of compression (default is (0, 1)).
    sym : bool, optional
        A flag to invoke a symmetry boundary in the direction of the thickness axis.

    Returns
    -------
    dict of felupe.Boundary
        Dict of boundaries for a biaxial loadcase.
    dict of ndarray
        Loadcase-related partitioned prescribed ``dof0`` and active ``dof1`` degrees of
        freedom as well as the external displacement values ``ext0`` for the prescribed
        degrees of freedom.

    Examples
    --------
    A rectangular planar specimen of a hyperelastic solid is subjected to a
    displacement-controlled combined shear-compression loading on rigid end faces.

    >>> import felupe as fem

    >>> mesh = fem.Rectangle(a=(0, 0), b=(4, 1), n=(41, 11))
    >>> region = fem.RegionQuad(mesh)
    >>> field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])

    The top edge is moved by ``-0.1`` to add a 10% constant compressive loading.

    >>> boundaries = fem.dof.shear(field, moves=(0, 0, -0.1))[0]

    The shear displacement is applied incrementally.

    >>> solid = fem.SolidBodyNearlyIncompressible(fem.NeoHooke(mu=1), field, bulk=5000)
    >>> step = fem.Step(
    >>>     items=[solid],
    >>>     ramp={boundaries["move"]: fem.math.linsteps([0, 1], num=5)},
    >>>     boundaries=boundaries
    >>> )

    >>> fem.Job(steps=[step]).evaluate()
    >>> img = field.screenshot("Principal Values of Logarithmic Strain")

    ..  image:: images/loadcase_shear.png

    See Also
    --------
    felupe.Boundary : A collection of prescribed degrees of freedom.
    felupe.dof.partition : Partition degrees of freedom into prescribed and active dof.
    felupe.dof.apply : Apply prescribed values for a list of boundaries.
    felupe.dof.symmetry : Return a dict of boundaries for the symmetry axes.

    """

    if move is not None:
        from warnings import warn

        warn(
            "The `move` argument is deprecated, use `moves` instead.",
            DeprecationWarning,
        )
        moves = list(moves)
        moves[0] = move

    if axis_compression is not None:
        from warnings import warn

        warn(
            "The `axis_compression` argument is deprecated, use `axes` instead.",
            DeprecationWarning,
        )
        axes = list(axes)
        axes[1] = axis_compression

    if axis_shear is not None:
        from warnings import warn

        warn(
            "The `axis_shear` argument is deprecated, use `axes` instead.",
            DeprecationWarning,
        )
        axes = list(axes)
        axes[0] = axis_shear

    if compression is not None:
        from warnings import warn

        warn(
            "The `compression` argument is deprecated, use `moves` instead.",
            DeprecationWarning,
        )
        moves = list(moves)
        moves[1] = compression[0]
        moves[2] = -compression[1]

    f = _get_first_field(field)

    if bottom is None:
        bottom = f.region.mesh.points[:, axes[1]].min()

    if top is None:
        top = f.region.mesh.points[:, axes[1]].max()

    if sym:
        sym = np.ones(3, dtype=bool)
        sym[axes,] = False

        bounds = symmetry(f, axes=sym)
    else:
        bounds = {}

    fy = ["fx", "fy", "fz"][axes[1]]

    skip_compression = [0, 0, 0]
    skip_compression[axes[1]] = 1

    not_skip_compression = [1, 1, 1]
    not_skip_compression[axes[1]] = 0

    not_skip_thickness = [0, 0, 0]
    not_skip_thickness[axes[1]] = 1
    not_skip_thickness[axes[0]] = 1

    not_skip_shear = [1, 1, 1]
    not_skip_shear[axes[0]] = 0

    bounds["bottom"] = Boundary(f, **{fy: bottom}, skip=skip_compression)
    bounds["top"] = Boundary(f, **{fy: top}, skip=not_skip_thickness)
    bounds["compression_bottom"] = Boundary(
        f, **{fy: bottom}, skip=not_skip_compression, value=moves[1]
    )
    bounds["compression_top"] = Boundary(
        f, **{fy: top}, skip=not_skip_compression, value=moves[2]
    )
    bounds["move"] = Boundary(f, **{fy: top}, skip=not_skip_shear, value=moves[0])

    dof0, dof1 = partition(field, bounds)
    ext0 = apply(field, bounds, dof0)

    return bounds, dict(dof0=dof0, dof1=dof1, ext0=ext0)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._base import Element


class Line(Element):
    r"""A 1D line element formulation with linear shape functions.

    Notes
    -----
    The linear line element is defined by two points (0-1). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r)`.

    .. math::

       \boldsymbol{h}(r) = \frac{1}{2} \begin{bmatrix}
               (1-r) \\
               (1+r)
           \end{bmatrix}

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self):
        super().__init__(shape=(2, 1))
        self.points = np.array([-1, 1], dtype=float)
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "line"

    def function(self, rv):
        "Return the shape functions at given coordinates (r,)."
        (r,) = rv
        return np.array([(1 - r), (1 + r)]) * 0.5

    def gradient(self, rv):
        "Return the gradient of shape functions at given coordinates (r,)."
        (r,) = rv
        return np.array([[-1], [1]]) * 0.5



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._models_hyperelasticity import NeoHooke
from ._models_linear_elasticity import lame_converter


class LinearElasticLargeStrain:
    r"""Linear-elastic material formulation suitable for large-rotation analyses based
    on the nearly-incompressible Neo-Hookean material formulation.

    Arguments
    ---------
    E : float
        Young's modulus.
    nu : float
        Poisson ratio.

    See Also
    --------
    NeoHooke: Nearly-incompressible isotropic hyperelastic Neo-Hooke material
        formulation.

    """

    def __init__(self, E=None, nu=None, parallel=False):
        self.E = E
        self.nu = nu

        # aliases for gradient and hessian
        self.energy = self.function
        self.stress = self.gradient
        self.elasticity = self.hessian

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [np.eye(3), np.zeros(0)]

        mu = None
        bulk = None

        if self.E is not None and self.nu is not None:
            gamma, mu = lame_converter(E, nu)
            bulk = gamma + 2 * mu / 3

        self.material = NeoHooke(mu=mu, bulk=bulk, parallel=parallel)

    def function(self, x, E=None, nu=None):
        """Evaluate the strain energy (as a function of the deformation gradient).

        Arguments
        ---------
        x : list of ndarray
            List with Deformation gradient ``F`` (3x3) as first item
        E : float, optional
            Young's modulus (default is None)
        nu : float, optional
            Poisson ratio (default is None)

        Returns
        -------
        ndarray
            Stress tensor (3x3)

        """

        if E is None:
            E = self.E

        if nu is None:
            nu = self.nu

        gamma, mu = lame_converter(E, nu)
        bulk = gamma + 2 * mu / 3

        return self.material.function(x, mu=mu, bulk=bulk)

    def gradient(self, x, E=None, nu=None):
        """Evaluate the stress tensor (as a function of the deformation gradient).

        Arguments
        ---------
        x : list of ndarray
            List with Deformation gradient ``F`` (3x3) as first item
        E : float, optional
            Young's modulus (default is None)
        nu : float, optional
            Poisson ratio (default is None)

        Returns
        -------
        ndarray
            Stress tensor (3x3)

        """

        if E is None:
            E = self.E

        if nu is None:
            nu = self.nu

        gamma, mu = lame_converter(E, nu)
        bulk = gamma + 2 * mu / 3

        return self.material.gradient(x, mu=mu, bulk=bulk)

    def hessian(self, x, E=None, nu=None):
        """Evaluate the elasticity tensor (as a function of the deformation gradient).

        Arguments
        ---------
        x : list of ndarray
            List with Deformation gradient ``F`` (3x3) as first item.
        E : float, optional
            Young's modulus (default is None)
        nu : float, optional
            Poisson ratio (default is None)

        Returns
        -------
        ndarray
            elasticity tensor (3x3x3x3)

        """

        if E is None:
            E = self.E

        if nu is None:
            nu = self.nu

        gamma, mu = lame_converter(E, nu)
        bulk = gamma + 2 * mu / 3

        return self.material.hessian(x, mu=mu, bulk=bulk)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from threading import Thread

import numpy as np

from .._cartesian import IntegralFormCartesian


class LinearForm:
    r"""A linear form object with methods for integration and assembly of
    vectors.

    ..  math::

        L(v) = \int_\Omega f \cdot v \ dx

    Parameters
    ----------
    v : BasisField
        An object with basis functions (gradients) of a field.
    grad_v : bool, optional (default is False)
        Flag to use the gradient of ``v``.
    dx : ndarray or None, optional (default is None)
        Array with (numerical) differential volumes.

    """

    def __init__(self, v, grad_v=False, dx=None):
        self.v = v
        self.grad_v = grad_v
        self.dx = dx

        self._form = IntegralFormCartesian(
            fun=None, v=v.field, dV=self.dx, grad_v=grad_v
        )

    def integrate(self, weakform, args=(), kwargs={}, parallel=False):
        r"""Return evaluated (but not assembled) integrals.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, *args, **kwargs)``.
        args : tuple, optional
            Optional arguments for callable weakform
        kawargs : dict, optional
            Optional named arguments for callable weakform
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.

        Returns
        -------
        values : ndarray
            Integrated (but not assembled) vector values.
        """

        if self.grad_v:
            v = self.v.grad
        else:
            v = self.v.basis

        values = np.zeros((len(v), *v.shape[-3:]))

        if not parallel:
            for a, vbasis in enumerate(v):
                for i, vb in enumerate(vbasis):
                    values[a, i] = weakform(vb, *args, **kwargs) * self.dx

        else:
            idx_a, idx_i = np.indices(values.shape[:2])
            ai = zip(idx_a.ravel(), idx_i.ravel())

            def contribution(values, a, i, args, kwargs):
                values[a, i] = weakform(v[a, i], *args, **kwargs) * self.dx

            threads = [
                Thread(target=contribution, args=(values, a, i, args, kwargs))
                for a, i in ai
            ]

            for t in threads:
                t.start()

            for t in threads:
                t.join()

        return values.sum(-2)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy as copy
from string import ascii_lowercase as alphabet

import numpy as np
from scipy.special import factorial

from ._base import Element


class ArbitraryOrderLagrange(Element):
    r"""A n-dimensional Lagrange finite element (e.g. line, quad or hexahdron) of
    arbitrary order.

    Notes
    -----

    **Polynomial shape functions**

    The basis function vector is generated with row-stacking of the individual lagrange
    polynomials. Each polynomial defined in the interval :math:`[-1,1]` is a function of
    the parameter :math:`r`. The curve parameters matrix :math:`\boldsymbol{A}` is of
    symmetric shape due to the fact that for each evaluation point :math:`r_j` exactly
    one basis function :math:`h_j(r)` is needed.

    ..  math::

        \boldsymbol{h}(r) = \boldsymbol{A}^T \boldsymbol{r}(r)

    **Curve parameter matrix**

    The evaluation of the curve parameter matrix :math:`\boldsymbol{A}` is carried out
    by boundary conditions. Each shape function :math:`h_i` has to take the value of one
    at the associated nodal coordinate :math:`r_i` and zero at all other point
    coordinates.

    ..  math::

        \boldsymbol{A}^T \boldsymbol{R} &=
            \boldsymbol{I} \qquad \text{with} \qquad \boldsymbol{R} =
            \begin{bmatrix}\boldsymbol{r}(r_1) & \boldsymbol{r}(r_2) & \dots &
            \boldsymbol{r}(r_p)\end{bmatrix}

        \boldsymbol{A}^T &= \boldsymbol{R}^{-1}


    **Interpolation and partial derivatives**

    The approximation of nodal unknowns :math:`\hat{\boldsymbol{u}}` as a function of
    the parameter :math:`r` is evaluated as

    ..  math::

        \boldsymbol{u}(r) \approx \hat{\boldsymbol{u}}^T \boldsymbol{h}(r)

    For the calculation of the partial derivative of the interpolation field w.r.t. the
    parameter :math:`r` a simple shift of the entries of the parameter vector is enough.
    This shifted parameter vector is denoted as :math:`\boldsymbol{r}^-`. A minus
    superscript indices the negative shift of the vector entries by :math:`-1`.

    ..  math::

        \frac{\partial \boldsymbol{u}(r)}{\partial r} &\approx
            \hat{\boldsymbol{u}}^T \frac{\partial \boldsymbol{h}(r)}{\partial r}

        \frac{\partial \boldsymbol{h}(r)}{\partial r} &=
            \boldsymbol{A}^T \boldsymbol{r}^-(r) \qquad \text{with} \qquad r_0^- =
            0 \qquad \text{and} \qquad r_i^- = \frac{r^{(i-1)}}{(i-1)!} \qquad
            \text{for} \qquad  i=(1 \dots p)


    n-dimensional shape functions
    *****************************

    Multi-dimensional shape function matrices
    :math:`\boldsymbol{H}_{2D}, \boldsymbol{H}_{3D}` are simply evaluated as dyadic
    (outer) vector products of one-dimensional shape function vectors. The multi-
    dimensional shape function vector is a one-dimensional representation (flattened
    version) of the multi-dimensional shape function matrix.

    ..  math::

        \boldsymbol{H}_{2D}(r,s) &= \boldsymbol{h}(r) \otimes \boldsymbol{h}(s)

        \boldsymbol{H}_{3D}(r,s,t) &= \boldsymbol{h}(r) \otimes \boldsymbol{h}(s)
            \otimes \boldsymbol{h}(t)
    """

    def __init__(self, order, dim, interval=(-1, 1)):
        self._order = order
        self._nshape = order + 1
        self._npoints = self._nshape**dim
        self._nbasis = self._npoints
        self._interval = interval

        super().__init__(shape=(self._npoints, dim))

        # init curve-parameter matrix
        n = self._nshape
        self._AT = np.linalg.inv(
            np.array([self._polynomial(p, n) for p in self._points(n)])
        ).T

        # indices for outer product in einstein notation
        # idx = ["a", "b", "c", ...][:dim]
        # subscripts = "a,b,c -> abc"
        self._idx = [letter for letter in alphabet][: self.dim]
        self._subscripts = ",".join(self._idx) + "->" + "".join(self._idx)

        # init points
        grid = np.meshgrid(*np.tile(self._points(n), (dim, 1)), indexing="ij")[::-1]
        self.points = np.vstack([point.ravel() for point in grid]).T

    def function(self, r):
        "Return the shape functions at given coordinate vector r."
        n = self._nshape

        # 1d - basis function vectors per axis
        h = [self._AT @ self._polynomial(ra, n) for ra in r]

        return np.einsum(self._subscripts, *h).ravel("F")

    def gradient(self, r):
        "Return the grgadient of shape functions at given coordinate vector r."
        n = self._nshape

        # 1d - basis function vectors per axis
        h = [self._AT @ self._polynomial(ra, n) for ra in r]

        # shifted 1d - basis function vectors per axis
        k = [self._AT @ np.append(0, self._polynomial(ra, n)[:-1]) for ra in r]

        # init output
        dhdr = np.zeros((n**self.dim, self.dim))

        # loop over columns
        for i in range(self.dim):
            g = copy(h)
            g[i] = k[i]
            dhdr[:, i] = np.einsum(self._subscripts, *g).ravel("F")

        return dhdr

    def _points(self, n):
        "Equidistant n-points in interval [-1, 1]."
        return np.linspace(*self._interval, n)

    def _polynomial(self, r, n):
        "Lagrange-Polynomial of order n evaluated at coordinate vector r."
        m = np.arange(n)
        return r**m / factorial(m)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from threading import Thread

import numpy as np

from .._cartesian import IntegralFormCartesian


class BilinearForm:
    r"""A bilinear form object with methods for integration and assembly of
    matrices.

    ..  math::

        a(v, u) = \int_\Omega v \cdot f \cdot u \ dx

    Parameters
    ----------
    v : BasisField
        An object with basis functions (gradients) of a field.
    grad_v : bool, optional (default is False)
        Flag to use the gradient of ``v``.
    u : BasisField
        An object with basis function (gradients) of a field.
    grad_u : bool, optional (default is False)
        Flag to use the gradient of ``u``.
    dx : ndarray or None, optional (default is None)
        Array with (numerical) differential volumes.

    """

    def __init__(self, v, u, grad_v=False, grad_u=False, dx=None):
        self.v = v
        self.grad_v = grad_v
        self.u = u
        self.grad_u = grad_u
        self.dx = dx

        self._form = IntegralFormCartesian(
            None, v.field, self.dx, u.field, grad_v, grad_u
        )

    def integrate(self, weakform, args=(), kwargs={}, parallel=False, sym=False):
        r"""Return evaluated (but not assembled) integrals.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, *args, **kwargs)``.
        args : tuple, optional
            Optional arguments for callable weakform
        kwargs : dict, optional
            Optional named arguments for callable weakform
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.
        sym : bool, optional (default is False)
            Flag to active symmetric integration/assembly.

        Returns
        -------
        values : ndarray
            Integrated (but not assembled) matrix values.
        """

        if self.grad_v:
            v = self.v.grad
        else:
            v = self.v.basis

        if self.grad_u:
            u = self.u.grad
        else:
            u = self.u.basis

        values = np.zeros((len(v), v.shape[-4], len(u), u.shape[-4], *u.shape[-2:]))

        if not parallel:
            for a, vbasis in enumerate(v):
                for i, vb in enumerate(vbasis):
                    for b, ubasis in enumerate(u):
                        for j, ub in enumerate(ubasis):
                            if sym:
                                if len(vbasis) * a + i <= len(ubasis) * b + j:
                                    values[a, i, b, j] = values[b, j, a, i] = (
                                        weakform(vb, ub, *args, **kwargs) * self.dx
                                    )

                            else:
                                values[a, i, b, j] = (
                                    weakform(vb, ub, *args, **kwargs) * self.dx
                                )

        else:
            idx_a, idx_i, idx_b, idx_j = np.indices(values.shape[:4])
            aibj = zip(idx_a.ravel(), idx_i.ravel(), idx_b.ravel(), idx_j.ravel())

            if sym:
                len_vbasis = values.shape[1]
                len_ubasis = values.shape[3]

                mask = len_vbasis * idx_a + idx_i <= len_ubasis * idx_b + idx_j
                idx_a, idx_i, idx_b, idx_j = (
                    idx_a[mask],
                    idx_i[mask],
                    idx_b[mask],
                    idx_j[mask],
                )

            def contribution(values, a, i, b, j, sym, args, kwargs):
                if sym:
                    values[a, i, b, j] = values[b, j, a, i] = (
                        weakform(v[a, i], u[b, j], *args, **kwargs) * self.dx
                    )

                else:
                    values[a, i, b, j] = (
                        weakform(v[a, i], u[b, j], *args, **kwargs) * self.dx
                    )

            threads = [
                Thread(
                    target=contribution, args=(values, a, i, b, j, sym, args, kwargs)
                )
                for a, i, b, j in aibj
            ]

            for t in threads:
                t.start()

            for t in threads:
                t.join()

        return values.sum(-2)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..assembly import IntegralForm
from ..constitution import AreaChange
from ..field import FieldAxisymmetric
from ..math import ddot, det, dot, dya, transpose
from ._helpers import Assemble, Evaluate, Results, StateNearlyIncompressible


class SolidBodyNearlyIncompressible:
    r"""A (nearly) incompressible SolidBody with methods for the assembly of sparse
    vectors/matrices for a material with optional state variables.

    Parameters
    ----------
    umat : A constitutive material formulation with methods for the evaluation
        of the gradient ``P = umat.gradient(F)`` as well as the hessian
        ``A = umat.hessian(F)`` of the strain energy function w.r.t. the
        deformation gradient.
    field : FieldContainer
        The field (and its underlying region) on which the solid body will
        be created on.
    bulk : float
        The bulk modulus of the volumetric material behaviour
        (:math:`U(J)=K(J-1)^2/2`).
    state : StateNearlyIncompressible
        A valid initial state for a (nearly) incompressible solid.

    Notes
    -----
    The volumetric material behaviour is hard-coded and is defined by the strain energy
    function.

    ..  math::

        U(J) = \frac{K}{2} (J - 1)^2


    **Hu-Washizu Three-Field-Variation Principle**

    The Three-Field-Variation :math:`(\boldsymbol{u},p,J)` leads to a linearized
    equation system with nine sub block-matrices. Due to the fact that the equation
    system is derived by a potential, the matrix is symmetric and hence, only six
    independent sub-matrices have to be evaluated. Furthermore, by the application of
    the mean dilatation technique, two of the remaining six sub-matrices are identified
    to be zero. That means four sub-matrices are left to be evaluated, where two
    non-zero sub-matrices are scalar-valued entries.

    ..  math::

        \begin{bmatrix}
            \boldsymbol{A}   & \boldsymbol{b} & \boldsymbol{0} \\
            \boldsymbol{b}^T &             0  &            -c  \\
            \boldsymbol{0}^T &            -c  &             d
        \end{bmatrix} \cdot \begin{bmatrix}
            \boldsymbol{x} \\
                        y  \\
                        z
        \end{bmatrix} = \begin{bmatrix}
            \boldsymbol{u} \\
                        v  \\
                        w
        \end{bmatrix}

    An alternative representation of the equation system, only dependent on the primary
    unknowns :math:`\boldsymbol{u}` is carried out. To do so, the second line is
    multiplied by :math:`\frac{d}{c}`.

    ..  math::

        \begin{bmatrix}
                        \boldsymbol{A}   & \boldsymbol{b} & \boldsymbol{0} \\
            \frac{d}{c}~\boldsymbol{b}^T &             0  &            -c  \\
                        \boldsymbol{0}^T &            -c  &             d
        \end{bmatrix} \cdot \begin{bmatrix}
            \boldsymbol{x} \\
                        y  \\
                        z
        \end{bmatrix} = \begin{bmatrix}
            \boldsymbol{u} \\
            \frac{d}{c}~v  \\
            -w
        \end{bmatrix}

    Now, equations two and three are summed up. This eliminates one of the three
    unknowns.

    ..  math::

        \begin{bmatrix}
                        \boldsymbol{A}   & \boldsymbol{b} \\
            \frac{d}{c}~\boldsymbol{b}^T &    -c
        \end{bmatrix} \cdot \begin{bmatrix}
            \boldsymbol{x} \\
                y
        \end{bmatrix} = \begin{bmatrix}
            \boldsymbol{u} \\
            \frac{d}{c}~v + w
        \end{bmatrix}

    Next, the second equation is left-multiplied by :math:`\frac{1}{c}~\boldsymbol{b}`
    and both equations are summed up again.

    ..  math::

        \begin{bmatrix}
            \boldsymbol{A} + \frac{d}{c^2}~\boldsymbol{b} \otimes \boldsymbol{b}
        \end{bmatrix} \cdot \begin{bmatrix}
            \boldsymbol{x}
        \end{bmatrix} = \begin{bmatrix}
            \boldsymbol{u} + \frac{d}{c^2}~\boldsymbol{b}~v +
                \frac{1}{c}~\boldsymbol{b}~w
        \end{bmatrix}

    The secondary unknowns are evaluated after solving the primary unknowns.

    ..  math::

        z &= \frac{1}{c}~\boldsymbol{b}^T \boldsymbol{x} - \frac{1}{c}~v

        y &= \frac{d}{c}~z - \frac{1}{c}~w

    For the mean-dilatation technique, the variables, equations as well as sub-matrices
    are evaluated. Note that the pairs of indices :math:`(ai)` and :math:`(bk)` have to
    be treated as 1d-vectors.

    ..  math::

        A_{aibk} &= \int_V \frac{\partial h_a}{\partial X_J}  \left(
            \frac{\partial^2 \overset{\wedge}{\psi}}{\partial F_{iJ} \partial F_{kL}} +
            p \frac{\partial^2 J}{\partial F_{iJ} \partial F_{kL}} \right)
            \frac{\partial h_b}{\partial X_L} \ dV

        b_{ai} &= \int_V \frac{\partial h_a}{\partial X_J}
            \frac{\partial J}{\partial F_{iJ}} \ dV

        c &= \int_V \ dV = V

        d &= \int_V \frac{\partial^2 U(\bar{J})}{\partial \bar{J} \partial \bar{J}} \ dV
           = \bar{U}'' V

    and

    ..  math::

        x_{ai} &= \delta {u}_{ai}

        y &= \delta p

        z &= \delta \bar{J}

    as well as

    ..  math::

        u_{ai} (= -r_{ai}) &= -\int_V \frac{\partial h_a}{\partial X_J} \left(
            \frac{\partial \overset{\wedge}{\psi}}{\partial F_{iJ}} +
            p \frac{\partial J}{\partial F_{iJ}} \right) \ dV

        v &= -\int_V (J - \bar{J}) \ dV = \bar{J} V - v

        w &= -\int_V (\bar{U}' - p) \ dV = p V - \bar{U}' V

    """

    def __init__(self, umat, field, bulk, state=None, statevars=None):
        self.umat = umat
        self.field = field
        self.bulk = bulk

        self._area_change = AreaChange()
        self._form = IntegralForm

        # volume of undeformed configuration
        if isinstance(self.field[0], FieldAxisymmetric):
            R = self.field[0].radius
            dA = self.field.region.dV
            dV = 2 * np.pi * R * dA
        else:
            dV = self.field.region.dV
        self.V = dV.sum(0)

        self.results = Results(stress=True, elasticity=True)

        if statevars is not None:
            self.results.statevars = statevars
        else:
            self.results.statevars = np.zeros(
                (
                    *umat.x[-1].shape,
                    field.region.quadrature.npoints,
                    field.region.mesh.ncells,
                )
            )

        if state is None:
            # init state of internal fields
            self.results.state = StateNearlyIncompressible(field)
        else:
            self.results.state = state

        self.results.kinematics = self._extract(self.field)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

        self.evaluate = Evaluate(
            gradient=self._gradient,
            hessian=self._hessian,
            cauchy_stress=self._cauchy_stress,
            kirchhoff_stress=self._kirchhoff_stress,
        )

    def _vector(self, field=None, parallel=False, items=None, args=(), kwargs={}):
        self.results.stress = self._gradient(
            field, parallel=parallel, args=args, kwargs=kwargs
        )

        form = self._form(
            fun=self.results.stress,
            v=self.field,
            dV=self.field.region.dV,
        )

        h = self.results.state.h(parallel=parallel)
        v = self.results.state.v()
        p = self.results.state.p

        values = [
            form.integrate(parallel=parallel)[0]
            + h * (self.bulk * (v / self.V - 1) - p)
        ]

        self.results.force = form.assemble(values=values)

        return self.results.force

    def _matrix(self, field=None, parallel=False, items=None, args=(), kwargs={}):
        self.results.elasticity = self._hessian(
            field, parallel=parallel, args=args, kwargs=kwargs
        )

        form = self._form(
            fun=self.results.elasticity,
            v=self.field,
            u=self.field,
            dV=self.field.region.dV,
        )

        h = self.results.state.h(parallel=parallel)

        values = [form.integrate(parallel=parallel)[0] + self.bulk / self.V * dya(h, h)]

        self.results.stiffness = form.assemble(values=values)

        return self.results.stiffness

    def _extract(self, field, parallel=False):
        u = field[0].values
        u0 = self.results.state.u
        h = self.results.state.h(parallel=parallel)
        v = self.results.state.v()
        J = self.results.state.J
        p = self.results.state.p

        du = (u - u0)[field.region.mesh.cells].transpose([1, 2, 0])

        # change of state variables due to change of displacement field
        dJ = ddot(h, du, mode=(2, 2)) / self.V + (v / self.V - J)
        dp = self.bulk * (dJ + J - 1) - p

        self.field = field
        self.results.kinematics = self.results.state.F = self.field.extract()

        # update state variables
        self.results.state.p = p + dp
        self.results.state.J = J + dJ
        self.results.state.u = u

        return self.results.kinematics

    def _gradient(self, field=None, parallel=False, args=(), kwargs={}):
        if field is not None:
            self.results.kinematics = self._extract(field, parallel=parallel)

        dJdF = self._area_change.function
        F = self.results.kinematics[0]
        statevars = self.results.statevars

        p = self.results.state.p

        gradient = self.umat.gradient([F, statevars], *args, **kwargs)

        self.results.stress = [gradient[0] + p * dJdF([F])[0]]
        self.results._statevars = gradient[-1]

        return self.results.stress

    def _hessian(self, field=None, parallel=False, args=(), kwargs={}):
        if field is not None:
            self.results.kinematics = self._extract(field, parallel=parallel)

        d2JdF2 = self._area_change.gradient
        F = self.results.kinematics[0]
        statevars = self.results.statevars
        p = self.results.state.p

        self.results.elasticity = [
            self.umat.hessian([F, statevars], *args, **kwargs)[0] + p * d2JdF2([F])[0]
        ]

        return self.results.elasticity

    def _kirchhoff_stress(self, field=None):
        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]

        return dot(P, transpose(F))

    def _cauchy_stress(self, field=None):
        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]
        J = det(F)

        return dot(P, transpose(F)) / J



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._boundary import Boundary


def get_dof0(field, bounds):
    "Extract prescribed degrees of freedom."

    # get mesh from field and obtain field-dimension
    mesh = field.region.mesh
    dim = field.dim

    # check if there are points without connected cells in the mesh
    # and add them to the list of prescribed dofs
    # e.g. these are points [2,6,7]
    #
    #   ( [[2,2,2], )   [[0,1,2],   [[ 6, 7, 8],
    # 3*(  [6,6,6], ) +  [0,1,2], =  [18,19,20],
    #   (  [7,7,7]] )    [0,1,2]]    [21,22,23]]
    #
    # fixmissing = [6, 7, 8, 18, 19, 29, 21, 22, 23]
    fixmissing = dim * np.tile(mesh.points_without_cells, (dim, 1)).T + np.arange(dim)

    # obtain prescribed dofs from boundaries
    dof0_bounds = np.concatenate([b.dof for b in bounds.values()])

    # combine all prescribed dofs and remove repeated itmes if there are any
    return np.unique(np.append(fixmissing.ravel(), dof0_bounds))


def get_dof1(field, bounds, dof0):
    "Extract active (non-prescribed) degrees of freedom."

    # obtain all dofs from the field
    dof = field.indices.dof

    # init a mask for the selection of active dofs
    mask = np.ones_like(dof.ravel(), dtype=bool)

    # set mask items for prescribed dofs (dof0) to False
    mask[dof0] = False

    # make the dof list 1d and mask active dofs
    return dof.ravel()[mask]


def partition(field, bounds):
    """Partition a list of degrees of freedom into prescribed (dof0) and active (dof1)
    degrees of freedom.

    Parameters
    ----------
    field : felupe.FieldContainer
        FieldContainer which holds the fields used in the boundaries.
    bounds : dict of felupe.Boundary
        Dict of boundaries.

    Returns
    -------
    dof0 : ndarray
        1d-array of int with all prescribed degress of freedom.
    dof1 : ndarray
        1d-array of int with all active degrees of freedom.

    Examples
    --------

    >>> import felupe as fem

    >>> mesh = fem.Rectangle(a=(0, 0), b=(1, 1), n=(3, 3))
    >>> region = fem.RegionQuad(mesh)
    >>> displacement = fem.FieldPlaneStrain(region, dim=2)
    >>> field = fem.FieldContainer([displacement])

    A plot shows the point-ids along with the associated degrees of freedom.

    >>> plotter = mesh.plot(off_screen=True)
    >>> plotter.add_point_labels(
    >>>     points=np.pad(mesh.points, ((0, 0), (0, 1))),
    >>>     labels=[
    >>>         f"Point {i}: DOF {a}, {b}"
    >>>         for i, (a, b) in enumerate(displacement.indices.dof)
    >>>     ],
    >>> )
    >>> img = plotter.screenshot("dof_partition.png", transparent_background=True)

    ..  image:: images/dof_partition.png

    >>> boundaries = dict(
    >>>     left=fem.Boundary(displacement, fx=0, value=0.2),
    >>>     right=fem.Boundary(displacement, fx=1),
    >>> )

    >>> dof0, dof1 = fem.dof.partition(field, boundaries)

    >>> dof0
    array([ 0,  1,  4,  5,  6,  7, 10, 11, 12, 13, 16, 17])

    >>> dof1
    array([ 2,  3,  8,  9, 14, 15])

    See Also
    --------
    felupe.Boundary : A collection of prescribed degrees of freedom.
    felupe.dof.apply : Apply prescribed values for a list of boundaries.
    """

    fields = field.fields
    offsets = field.offsets

    # list of boundaries, partitioned by fields
    boundaries = [
        {key: value for key, value in bounds.items() if value.field == f}
        for f in fields
    ]

    # fix fields without boundaries (need at least one boundary per field)
    boundaries = [
        {"__empty__": Boundary(f)} if not b else b for f, b in zip(fields, boundaries)
    ]

    # partition degrees of freedom for each field
    dofs0 = [get_dof0(f, b) for f, b in zip(fields, boundaries)]
    dofs1 = [get_dof1(f, b, dof0=i) for f, b, i in zip(fields, boundaries, dofs0)]

    # concatenate degrees of freedom
    dof0 = np.concatenate(
        [dof0 + offset for dof0, offset in zip(dofs0, np.insert(offsets, 0, 0))]
    )
    dof1 = np.concatenate(
        [dof1 + offset for dof1, offset in zip(dofs1, np.insert(offsets, 0, 0))]
    )

    return dof0, dof1


def apply(field, bounds, dof0=None):
    """Apply prescribed values for a list of boundaries and return all (default) or only
    the prescribed components of the ``field`` based on ``dof0``.

    Parameters
    ----------
    field : felupe.FieldContainer
        FieldContainer which holds the fields used in the boundaries.
    bounds : dict of felupe.Boundary
        Dict of boundaries.
    dof0 : ndarray or None, optional
        1d-array of int with prescribed degrees of freedom (default is None). If not
        None, only the given deegrees of freedom ``dof0`` of the field values,
        prescribed by the boundaries, are returned.

    Returns
    -------
    ndarray
        Field values at mesh-points for all (default) or only the prescribed components
        of the ``field`` based on ``dof0``.

    Examples
    --------

    >>> import felupe as fem

    >>> mesh = fem.Rectangle(a=(0, 0), b=(1, 1), n=(3, 3))
    >>> region = fem.RegionQuad(mesh)
    >>> displacement = fem.FieldPlaneStrain(region, dim=2)
    >>> field = fem.FieldContainer([displacement])

    >>> boundaries = dict(
    >>>     left=fem.Boundary(displacement, fx=0, value=0.2),
    >>>     right=fem.Boundary(displacement, fx=1),
    >>> )

    >>> dof0, dof1 = fem.dof.partition(field, boundaries)
    >>> ext0 = fem.dof.apply(field, boundaries, dof0=dof0)

    >>> dof0
    array([ 0,  1,  4,  5,  6,  7, 10, 11, 12, 13, 16, 17])

    >>> dof1
    array([ 2,  3,  8,  9, 14, 15])

    >>> ext0
    array([0.2, 0.2, 0. , 0. , 0.2, 0.2, 0. , 0. , 0.2, 0.2, 0. , 0. ])

    ``dof0=None`` is required (default) if the prescribed displacement array should be
    returned for all degrees of freedom.

    >>> fem.dof.apply(field, boundaries).reshape(
    >>>     displacement.values.shape
    >>> )
    array([[0.2, 0.2],
           [0. , 0. ],
           [0. , 0. ],
           [0.2, 0.2],
           [0. , 0. ],
           [0. , 0. ],
           [0.2, 0.2],
           [0. , 0. ],
           [0. , 0. ]])

    See Also
    --------
    felupe.Boundary : A collection of prescribed degrees of freedom.
    felupe.dof.partition : Partition degrees of freedom into prescribed and active dof.
    """

    # check if a mixed-field is passed
    u = np.concatenate([f.values.ravel() for f in field.fields])
    offsets = np.insert(field.offsets, 0, 0)

    for b in bounds.values():
        # get offset for field-dof of current boundary
        offset = offsets[[b.field == f for f in field.fields]]

        # set prescribed values
        u.ravel()[b.dof + offset] = b.value

    if dof0 is None:
        return u
    else:
        return u.ravel()[dof0]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._base import Element
from ._lagrange import ArbitraryOrderLagrange


class ConstantQuad(Element):
    r"""A 2D quadrilateral element formulation with constant shape functions.

    Notes
    -----
    The quadrilateral element is defined by four points (0-3). [1]

    The shape function :math:`h` is given in terms of the coordinates :math:`(r,s)`.

    .. math::

       h(r,s) = 1

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.ConstantQuad().screenshot()

    ..  image:: images/quad.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self):
        super().__init__(shape=(1, 2))
        self.points = np.array([[-1, -1], [1, -1], [1, 1], [-1, 1]], dtype=float)
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "quad"

    def function(self, rs):
        "Return the shape functions at given coordinates (r, s)."
        return np.array([1])

    def gradient(self, rs):
        "Return the gradient of shape functions at given coordinates (r, s)."
        return np.array([[0, 0]])


class Quad(Element):
    r"""A 2D quadrilateral element formulation with linear shape functions.

    Notes
    -----
    The quadrilateral element is defined by four points (0-3). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s)`.

    .. math::

       \boldsymbol{h}(r,s) = \frac{1}{4} \begin{bmatrix}
               (1-r) (1-s) \\
               (1+r) (1-s) \\
               (1+r) (1+s) \\
               (1-r) (1+s)
           \end{bmatrix}

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.Quad().screenshot()

    ..  image:: images/quad.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self):
        super().__init__(shape=(4, 2))
        self.points = np.array([[-1, -1], [1, -1], [1, 1], [-1, 1]], dtype=float)
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "quad"

    def function(self, rs):
        "Return the shape functions at given coordinates (r, s)."
        r, s = rs
        return (
            np.array(
                [
                    (1 - r) * (1 - s),
                    (1 + r) * (1 - s),
                    (1 + r) * (1 + s),
                    (1 - r) * (1 + s),
                ]
            )
            * 0.25
        )

    def gradient(self, rs):
        "Return the gradient of shape functions at given coordinates (r, s)."

        r, s = rs
        return (
            np.array(
                [
                    [-(1 - s), -(1 - r)],
                    [(1 - s), -(1 + r)],
                    [(1 + s), (1 + r)],
                    [-(1 + s), (1 - r)],
                ]
            )
            * 0.25
        )


class QuadraticQuad(Element):
    r"""A 2D quadrilateral element formulation with quadratic (serendipity) shape
    functions.

    Notes
    -----
    The quadratic (serendipity) quadrilateral element is defined by eight points (0-7).
    [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s)`.

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.QuadraticQuad().screenshot()

    ..  image:: images/quad8.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self):
        super().__init__(shape=(8, 2))
        self.points = np.array(
            [
                [-1, -1],
                [1, -1],
                [1, 1],
                [-1, 1],
                [0, -1],
                [1, 0],
                [0, 1],
                [-1, 0],
            ],
            dtype=float,
        )
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "quad8"

    def function(self, rs):
        "Return the shape functions at given coordinates (r, s)."
        r, s = rs
        ra, sa = self.points.T

        h = (1 + ra * r) * (1 + sa * s) * (ra * r + sa * s - 1) / 4
        h[ra == 0] = (1 - r**2) * (1 + sa[ra == 0] * s) / 2
        h[sa == 0] = (1 + ra[sa == 0] * r) * (1 - s**2) / 2

        return h

    def gradient(self, rs):
        "Return the gradient of shape functions at given coordinates (r, s)."

        r, s = rs
        ra, sa = self.points.T

        dhdr = (
            ra * (1 + sa * s) * (ra * r + sa * s - 1) / 4
            + (1 + ra * r) * (1 + sa * s) * ra / 4
        )

        dhdr[ra == 0] = -2 * r * (1 + sa[ra == 0] * s) / 2
        dhdr[sa == 0] = ra[sa == 0] * (1 - s**2) / 2

        dhds = (1 + ra * r) * sa * (ra * r + sa * s - 1) / 4 + (1 + ra * r) * (
            1 + sa * s
        ) * sa / 4

        dhds[ra == 0] = (1 - r**2) * sa[ra == 0] / 2
        dhds[sa == 0] = (1 + ra[sa == 0] * r) * -2 * s / 2

        return np.vstack([dhdr, dhds]).T


class BiQuadraticQuad(Element):
    r"""A 2D quadrilateral element formulation with bi-quadratic shape functions.

    Notes
    -----
    The bi-quadratic quadrilateral element is defined by nine points (0-8). [1]

    The shape functions :math:`\boldsymbol{h}` are given in terms of the coordinates
    :math:`(r,s)`.

    Examples
    --------
    >>> import felupe as fem

    >>> img = fem.BiQuadraticQuad().screenshot()

    ..  image:: images/quad9.png

    References
    ----------
    .. [1] W. Schroeder, K. Martin and B. Lorensen. The Visualization
       Toolkit, 4th ed. Kitware, 2006. ISBN: 978-1-930934-19-1.
    """

    def __init__(self):
        super().__init__(shape=(9, 2))

        self._lagrange = ArbitraryOrderLagrange(order=2, dim=2)

        self._vertices = np.array([0, 2, 8, 6])
        self._edges = np.array([1, 5, 7, 3])
        self._faces = np.array([4])
        self._volume = np.array([], dtype=int)

        self._permute = np.concatenate(
            (self._vertices, self._edges, self._faces, self._volume)
        )

        self.points = self._lagrange.points[self._permute]
        self.cells = np.arange(len(self.points)).reshape(1, -1)
        self.cell_type = "quad9"

    def function(self, rs):
        "Return the shape functions at given coordinates (r, s)."

        return self._lagrange.function(rs)[self._permute]

    def gradient(self, rs):
        "Return the gradient of shape functions at given coordinates (r, s)."

        return self._lagrange.gradient(rs)[self._permute, :]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ._basis import Basis
from ._mixed import BilinearFormExpression, LinearFormExpression


class FormExpression:
    r"""A linear or bilinear form object based on a weak-form with
    methods for integration and assembly of vectors / sparse matrices.

    Linear Form:

    ..  math::

        L(v) = \int_\Omega f \cdot v \ dx

    Bilinear Form:

    ..  math::

        a(v, u) = \int_\Omega v \cdot f \cdot u \ dx

    Parameters
    ----------
    v : Field or FieldMixed
        An object with interpolation or gradients of a field. May be
        updated during integration / assembly.
    u : Field or FieldMixed
        An object with interpolation or gradients of a field. May be
        updated during integration / assembly.
    grad_v : bool, optional
        Flag to use the gradient of ``v`` (default is False).
    grad_u : bool, optional
        Flag to use the gradient of ``u`` (default is False).
    dx : ndarray or None, optional
        Array with (numerical) differential volumes (default is None).
    args : tuple or None, optional
        Tuple with initial optional weakform-arguments. May be updated during
        integration / assembly (default is None).
    kwargs : dict or None, optional
        Dictionary with initial optional weakform-keyword-arguments. May be updated
        during integration / assembly (default is None).
    parallel : bool, optional
        Flag to activate parallel (threaded) basis evaluation (default is False).

    """

    def __init__(
        self,
        weakform,
        v,
        u=None,
        grad_v=False,
        grad_u=False,
        dx=None,
        args=None,
        kwargs=None,
        parallel=False,
    ):
        # set attributes
        self.form = None
        self.grad_u = grad_u
        self.grad_v = grad_v
        self.dx = dx
        self.weakform = weakform

        if args is not None:
            self.args = args
        else:
            self.args = ()

        if kwargs is not None:
            self.kwargs = kwargs
        else:
            self.kwargs = {}

        # init underlying linear or bilinear (mixed) form
        self._init_or_update_forms(v, u, args, kwargs, parallel)

    def _init_or_update_forms(self, v, u, args, kwargs, parallel):
        "Init or update the underlying form object."

        # update args for weakform
        if args is not None:
            self.args = args

        # update kwargs for weakform
        if kwargs is not None:
            self.kwargs = kwargs

        # get current form type
        if self.form is not None:
            form_type = type(self.form)
        else:
            form_type = None

        if v is not None:
            # linear form
            if u is None:
                self.u = None

                # mixed-field input
                self.v = Basis(v, parallel=parallel)
                form = LinearFormExpression(self.v, self.grad_v)

                # evaluate weakform to list of weakforms
                if isinstance(self.weakform, type(lambda x: x)):
                    self.weakform = self.weakform()

            else:
                self.v = Basis(v, parallel=parallel)
                self.u = Basis(u, parallel=parallel)
                form = BilinearFormExpression(self.v, self.u, self.grad_v, self.grad_u)

                # evaluate weakform to list of weakforms
                if isinstance(self.weakform, type(lambda x: x)):
                    self.weakform = self.weakform()

            # check if new form type matches initial form type (update-stage)
            if form_type is not None:
                if form_type == type(form):
                    self.form = form
                else:
                    raise TypeError("Wrong type of fields.")

            # set form without a check (init-stage)
            else:
                self.form = form

    def integrate(
        self, v=None, u=None, args=None, kwargs=None, parallel=False, sym=False
    ):
        r"""Return evaluated (but not assembled) integrals.

        Parameters
        ----------
        v : Field, FieldMixed or None, optional
            An object with interpolation or gradients of a field as specified
            by boolean flag ``self.grad_v`` (default is None).
        u : Field, FieldMixed or None, optional
            An object with interpolation or gradients of a field as specified
            by boolean flag ``self.grad_v`` (default is None).
        args : tuple, optional (default is ())
            Tuple with optional weakform-arguments.
        kwargs : dict, optional (default is {})
            Dictionary with optional weakform-keyword-arguments.
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.
        sym : bool, optional (default is False)
            Flag to active symmetric integration/assembly
            for bilinear forms.

        Returns
        -------
        values : ndarray
            Integrated (but not assembled) vector / matrix values.
        """

        self._init_or_update_forms(v, u, args, kwargs, parallel)

        kwargs = dict(parallel=parallel, sym=sym)

        if self.u is None:
            kwargs.pop("sym")

        return self.form.integrate(
            self.weakform, args=self.args, kwargs=self.kwargs, **kwargs
        )

    def assemble(
        self, v=None, u=None, args=None, kwargs=None, parallel=False, sym=False
    ):
        r"""Return the assembled integral as vector / sparse matrix.

        Parameters
        ----------
        v : Field, FieldMixed or None, optional
            An object with interpolation or gradients of a field as specified
            by boolean flag ``self.grad_v`` (default is None).
        u : Field, FieldMixed or None, optional
            An object with interpolation or gradients of a field as specified
            by boolean flag ``self.grad_v`` (default is None).
        args : tuple, optional (default is ())
            Tuple with optional weakform-arguments.
        kwargs : dict, optional (default is {})
            Dictionary with optional weakform-keyword-arguments.
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.
        sym : bool, optional (default is False)
            Flag to active symmetric integration/assembly
            for bilinear forms.

        Returns
        -------
        values : csr_matrix
            The assembled vector / sparse matrix.
        """

        self._init_or_update_forms(v, u, args, kwargs, parallel)

        kwargs = dict(parallel=parallel, sym=sym)

        if self.u is None:
            kwargs.pop("sym")

        return self.form.assemble(
            self.weakform, args=self.args, kwargs=self.kwargs, **kwargs
        )
